package rover;

//import gov.nasa.jpf.jvm.Verify;
import java.util.Vector;

import com.ibm.contest.utils.QuickRandom;

class AbstractExecutive extends Executive {

	public QuickRandom ctRandom = QuickRandom.newInstance((int) System.currentTimeMillis());
	
	public static final int SUCCESS = 1;

	public static final int FAILURE = -1;

	public static final int START = 0;

	public AbstractExecutive() {
		super();
	}

	public AbstractExecutive(CommandTable cmdtab, FunctionTable fntab,
			ExternalSystem externSys, ExecDisplay xdisplay,
			TelemetryInterface xtelemInterface) {
		//Verify.beginAtomic();
		initialize();
		plan = lookForNewPlan();

		db = new Database();
		conditionChecker = new ExecConditionChecker(db, this, null);
		timer = Factory.newExecTimer();

		timerWaiter = new ExecTimerWaiter(this, timer);
		actionThread = new ActionExecution(this, externSys, null);

		commandTable = cmdtab;
		functionTable = fntab;
		//Verify.endAtomic();
	}

	void condition(Node cn, int i) {
		// trap these results to create properties
	}

	// topLevelLoop
	public void run() {
		synchronized (this) {
			//Verify.beginAtomic();
			boolean skip_once = true;
			boolean terminate = false;
			while (!terminate) {
				try {
					// Look for new plan.
					if (debug)
						System.out
								.println("Executive: run - check for new plan");
					
					if (skip_once) { // so that we only read one plan
						skip_once = false;
					} else {
						while (pendingPlanFile == null) { // Wait for a change.
							if (debug)
								System.out
										.println("Executive: run - no plan wait");
							//Verify.endAtomic();
							// don't want JPF to wait here since 
							// this deadlock spoils the fun
							//ZL Verify.ignoreIf(true);
							while (true) {
								skip_once = false;
							}
						}
					}
					
					if (debug)
						System.out.println("Executive: run - plan signaled");
					resetExecutionSignal();
					pendingPlanFile = null;
					if (plan != null) {
						//Verify.endAtomic();
						changePlan(plan, PlanInstallType.PlanInstall_Replace);
						//Verify.beginAtomic();
					} else {
						System.out.println("Got pending file, but no plan");
						pendingPlanFile = null;
					}
					terminate = true;
				} catch (ExecutionException except) {
					switch (except.getException()) {
					case ExecutionException.ExecTerminate:
						terminate = true;
						break;
					default:
						System.out
								.println("Unknown ExecutionException at top level"
										+ except);
					}
				}
			}
			// We should never get here unless we're done.
			// TODO: unlock the mutex in the thread cleanup code.
		}
	}

	int actionResult(Task current_task, DBValue dbval) {
//ZL		int result = Verify.random(2);
		int result = ctRandom.nextInt(2);
		switch (result) {
		case 0:
			return ActionStatus.ActionSuccess;
		case 1:
			return ActionStatus.ActionFailure;
		default:
			return ActionStatus.ExecTerminateStatus;
		}
	}

	boolean isSuccessReturnValue(int retval) {
//ZL		return Verify.randomBool();
		return ctRandom.nextBoolean();
	}

	boolean isFailureReturnValue(int retval) {
//ZL		return Verify.randomBool();
		return ctRandom.nextBoolean();
	}

	boolean isTerminateReturnValue(int retval) {
//ZL		return Verify.randomBool();
		return ctRandom.nextBoolean();
	}

	void optionalAssert() {
		db.DBAssert();
	}
}
