package rover;

import com.ibm.contest.utils.QuickRandom;
//import gov.nasa.jpf.jvm.Verify;

class ActionExecution extends Thread {

	public QuickRandom ctRandom = QuickRandom.newInstance((int) System.currentTimeMillis()); 
	
  public static final int EXEC_RETURNVALUE_UNKNOWN  = -1;
  public static final int EXEC_RETURNVALUE_PENDING  =  0;
  public static final int EXEC_RETURNVALUE_SUCCESS  =  1;
  public static final int EXEC_RETURNVALUE_FAILURE  =  2;
  public static final int EXEC_RETURNVALUE_TERMINATE = 666;

  public static final int ACTION_STATUS_POLL_MSECS  = 200;

  class timespec {
    long tv_sec;
    long tv_nsec;
  }

  class timeval {
    long tv_sec;
    long tv_usec;
  }


  public static final boolean debug = Main.print;

  Executive exec;
  ExternalSystem externSys;

  // Halted == waiting for execution to halt.
  // Pending == waiting for execution to start.
  public static final int Running = 0;
  public static final int Idle    = 1;
  public static final int Halted  = 2;
  public static final int Pending = 3;

  int /*ExecutionStatus*/ status;

  Symbol current_cmdFnName;
  int /*ExecCmdIDType*/ current_cmdID;
  boolean current_asynchronousExecution;
  Symbol current_interruptCmdFnName;
  int /*ExecCmdIDType*/ current_interruptCmdID;
  boolean current_interruptAsynchronousExecution;

  // executing_cmdID is the thread's current cmdID, which could
  // be different than current_cmdID if a new action has been
  // proposed but not yet picked up on by the thread.
  int /*ExecCmdIDType*/ executing_cmdID;

  public ActionExecution(Executive xexec, ExternalSystem xexternSys,
			 ExecHasThreadedComponents xparent) {
    exec = xexec;
    externSys = xexternSys;
    status = Idle;
    current_cmdID = 0;
    current_interruptCmdID = 0;
    current_cmdFnName = Symbol.NullSymbol; 
    current_interruptCmdFnName = Symbol.NullSymbol;
    start();
  }

  public void acknowledgeCondVarSignal() {
    notifyAll();
  }

  public void notifyCondVarSignal() {
    notifyAll();
  }

  public void notifyCondVarWait() {
    try{wait();}catch(InterruptedException e){};
  }

  public void acknowledgeCondVarWait() {
    try{wait();}catch(InterruptedException e){};
  }

  boolean notifyCondVarTimedWait(timespec to) { //WV untimed wait
    if (Main.JPF) { // WV move to factory!
        return ctRandom.nextBoolean();
//ZL      return Verify.randomBool();
    }
    long wait_time = to.tv_sec*1000+to.tv_nsec/1000000;
    long milliseconds = System.currentTimeMillis();    
    long new_wait_time = wait_time - milliseconds;
    if (debug)
      System.out.println(Thread.currentThread()+"Wait for " + new_wait_time);
    if (new_wait_time > 0) 
      try{wait(new_wait_time);}catch(InterruptedException e){};
    if (debug)
      System.out.println(Thread.currentThread()+"Done Waiting");
    if (System.currentTimeMillis() > wait_time) 
      return true;
    else
      return false; // no timeout
  }

  void internalDoAction(Symbol cmdFnName,
			int /*WV ExecCmdIDType*/ cmdID,
			boolean asynchronousExecution) {
    if (debug)
      System.out.println(Thread.currentThread()+"ActionExecution: internalDoAction(" + cmdID + "," 
			 + asynchronousExecution);
    if (cmdFnName == null) {
      System.out.println(Thread.currentThread()+"ActionExecution: Null command function");
      // If there's no action, it automatically succeeds, although the
      // plan is malformed.
      // TODO: should we be throwing a plan syntax error at this point?
      synchronized(this) {
	exec.assertCommandStatus(cmdID, ActionStatus.ActionSuccess);
	status = Idle;
      }
      return;
    } else {
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: cmdFnName='" + cmdFnName);
    }
    if (asynchronousExecution) {
      // TODO: Call the function and then poll for action status.
      // Call function.
      try {
	externSys.executeAction();
      } catch (ExecutionError execError) {
	System.out.println(Thread.currentThread()+"Caught execution error " + execError);
	status = Idle;
	return;
      }
      if (debug)
      System.out.println(Thread.currentThread()+"ActionExecution: returned from calling command function");
      timespec to = new timespec();
      int commandStatus = EXEC_RETURNVALUE_UNKNOWN;
      synchronized(this) {
	for (;;) {
	  boolean condvar_retval;
	  timeval tp = new timeval();
	  long milliseconds = System.currentTimeMillis();
	  tp.tv_usec = (milliseconds % 1000) * 1000;
	  tp.tv_sec  = milliseconds/1000;
	  if (/*gettimeofday(&tp, 0)*/ milliseconds < 0) {
	    System.out.println(Thread.currentThread()+"gettimeofday: Error ");
	    break;
	  }
	  timeval sleep_to = new timeval();
	  sleep_to.tv_sec = tp.tv_sec;
	  sleep_to.tv_usec = tp.tv_usec + ACTION_STATUS_POLL_MSECS * 1000;
	  while (sleep_to.tv_usec > 1000000) {
	    sleep_to.tv_usec -= 1000000;
	    sleep_to.tv_sec++;
	  }
	  to.tv_sec = sleep_to.tv_sec;
	  to.tv_nsec = sleep_to.tv_usec * 1000;
	  do {
	    condvar_retval = notifyCondVarTimedWait(to);
	    // condvar_retval = true if ETIMEOUT
	  } while (!condvar_retval && status != Halted);
	  if (status == Halted) {
	    // Most important -- cancel if action is halted.
	    // Acknowledge receipt of signal.
	    
	    // TODO: abort action.

	    status = Idle;
	    acknowledgeCondVarSignal();
	    break;
	  } else if (condvar_retval) {
	    
	    // Poll execution status.
	    int newCommandStatus = externSys.getCommandStatus(/*WV cmdID*/);
	    if (commandStatus != newCommandStatus) {
	      // A change in status.
	      commandStatus = newCommandStatus;
	      if (debug)
		System.out.println(Thread.currentThread()+"ActionExecution: about to assert commandStatus " + commandStatus);
	      exec.assertCommandStatus(cmdID, commandStatus);
	      if (debug)
		System.out.println(Thread.currentThread()+"ActionExecution: just asserted commandStatus " + commandStatus);
	      // sleep(1);
	      if (commandStatus == EXEC_RETURNVALUE_SUCCESS ||
		  commandStatus == EXEC_RETURNVALUE_FAILURE) {
		if (debug)
		  System.out.println(Thread.currentThread()+"ActionExecution: got termination status " + commandStatus);
		status = Idle;
		break;
	      }
	    }
	  }
	}
      } 
    } else { // synchronous action
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: synchronous execution");
      int cmdStatus = EXEC_RETURNVALUE_UNKNOWN;
      try {
	cmdStatus = 
	  externSys.executeSynchronousAction(/*WV cmdFnName, 
					       cmdID, argvals,this*/ );
      } catch (ExecutionError execError) {
	System.out.println(Thread.currentThread()+"Caught execution error " + execError);
	status = Idle;
	return;
      }
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: got command status " + 
			   cmdStatus + " for cmdID " + cmdID);
      // TODO: update action status internally. (??)
      synchronized(this) {
	exec.assertCommandStatus(cmdID, cmdStatus);
	if (debug)
	  System.out.println(Thread.currentThread()+"ActionExecution: done asserting command status " +
			     cmdStatus + " for cmdID " + cmdID);
	status = Idle;
      }
    }
  }



  //runActionExecutionThreadInternal()
  public void run() {
    //Verify.beginAtomic();
    boolean doInternal = false;
    if (debug)
      System.out.println(Thread.currentThread()+"ActionExecution: starting thread");
    for (;;) {
      //Verify.endAtomic();
      synchronized(this) {
	//Verify.beginAtomic();
	if (debug)
	  System.out.println(Thread.currentThread()+"ActionExecution: got mutex");
	if (status == Pending) {
	  // Then there's something waiting.
	  if (debug)
	    System.out.println(Thread.currentThread()+"ActionExecution: status = Running");
	  status = Running;
	  executing_cmdID = current_cmdID;
	  // Acknowledge receipt of the action.
	  acknowledgeCondVarSignal();
	  if (debug)
	    System.out.println(Thread.currentThread()+"ActionExecution: signaled acknowledgeCondVar");
	  doInternal = true;
	} else if (status == Halted) {
	  System.out.println(Thread.currentThread()+"ActionExecution: Trying to halt non-running execution\n");
	  // Set the status to Idle to wait for the next action.
	  // TODO: what happens if we get the abort before an action starts?
	  status = Idle;
	} else {
	  if (debug)
	    System.out.println(Thread.currentThread()+"ActionExecution: waiting for notifyCondVar");
	  //Verify.endAtomic();
	  notifyCondVarWait();
	  //Verify.beginAtomic();
	}
      }
      if (doInternal) {
	//Verify.endAtomic();
	internalDoAction(/*current_cmdFnName*/ new Symbol("dummy"), 
			 current_cmdID, // actions always asynchronous
			 /*current_asynchronousExecution*/ true); 
	
	//Verify.beginAtomic();
	doInternal = false;
      }
    }
  }


  void stopActionBody() {
    if (status == Running) {
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: status = Halted");
      status = Halted;
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: notifying CondVar");
      notifyCondVarSignal();
      if (debug)
        System.out.println(Thread.currentThread()+"ActionExecution: waiting for acknowledge CondVar");
      while (status != Idle) {
	//Verify.endAtomic();
	acknowledgeCondVarWait();
        //Verify.beginAtomic();
      }
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: status = Idle");
      if (debug)
	System.out.println(Thread.currentThread()+"About to run interrupt action");
    }
  }

  void stopAction(boolean lock) {
    //Verify.beginAtomic();
    if (lock) {
      //Verify.endAtomic();
      synchronized(this) {
        //Verify.beginAtomic();
	stopActionBody();
      }
    }
    else {
      stopActionBody();
    }
    //Verify.endAtomic();
  }

  void abortActionBody() {
    if (debug)
      System.out.println(Thread.currentThread()+"ActionExecution: running abortAction()");
    if (status == Running) {
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: abortAction: about to stopAction");
      //Verify.endAtomic();
      stopAction(false); // Don't lock recursively.
      //Verify.beginAtomic();
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: abortAction: finished with stopAction");
      installAction();
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: abortAction: finished with installAction of interrupt command");
    }
  }

  void abortAction(boolean lock) {
    //Verify.beginAtomic();
    if (lock) {
      //Verify.endAtomic();
      synchronized(this) {
        //Verify.beginAtomic();
	abortActionBody();
      }
    }
    else {
      abortActionBody();
    }
    //Verify.endAtomic();
  }

  public void doAction(boolean lock) {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"ActionExecution: doAction: cmdID=");
    if (!lock) 
      System.out.println(Thread.currentThread()+"OOPS: Lock was supposed to hold: in doAction");
    //Verify.endAtomic();
    synchronized(this) {
      //Verify.beginAtomic();
      if (debug)
	System.out.println(Thread.currentThread()+"ActionExecution: doAction: got mutex");
      if (status == Running) {
	//Verify.endAtomic();
	stopAction(false); // Don't lock recursively.
        //Verify.beginAtomic();
	installAction();
      } else if (status == Halted)
	System.out.println(Thread.currentThread()+"ActionExecution: doAction called when in Halted state");
      else {
	if (debug)
	  System.out.println(Thread.currentThread()+"ActionExecution: doAction: calling installAction");
	installAction();
      }
    }
    //Verify.endAtomic();
  }

  public void installAction() { //WV  Heavily abstracted
    status = Pending;
    notifyCondVarSignal();
  }

}
