package rover;
//import gov.nasa.jpf.jvm.Verify;

import java.util.Vector;
import java.util.Iterator;

class ExecConditionChecker extends java.lang.Thread { 
  //WV inherits from MutexMultiCondVar<2> and ExecHasThreadedComponents

  static final boolean debug = Main.print;

  class ExecConditionCheckerThread extends java.lang.Thread {
    ExecConditionChecker parent;
    ExecHasThreadedComponents threadParent;

    public ExecConditionCheckerThread(ExecConditionChecker xparent,
				      ExecHasThreadedComponents xthreadParent){
      parent = xparent;
    }

    public void run() {
      parent.internalConditionCheckerLoop();
    }

  }

  class ExecDBMonitorThread extends java.lang.Thread {
    ExecConditionChecker parent;
    ExecHasThreadedComponents threadParent;

    public ExecDBMonitorThread(ExecConditionChecker xparent,
			       ExecHasThreadedComponents xthreadParent) {
      parent = xparent;
    }

    public void run() {
      parent.internalDBMonitorLoop();
    }
  }

  ExecConditionCheckerThread checkerThread;
  ExecDBMonitorThread dbThread;

  Database db;
  Executive exec;

  ExpListOfExecCCheckerStrs conditionSet = new  ExpListOfExecCCheckerStrs();
  boolean conditionSetChanged;
  boolean checkerDBChanged;
  static int conditionRecheckIndex; // = 0
  static int metstructsIndex; // = 1

  ExecConditionCheckerStruct savedWakeupStruct;

  ExecConditionCheckerStruct getSavedWakeup() { 
    return savedWakeupStruct; 
  }

  void recheckCondVarSignal() {
    notifyAll();
  }

  void recheckCondVarWait() {
    try{wait();}catch(InterruptedException e){};
  }

  void recheckCondVarTimedWait(/*WV timespec& to*/) {
    try{wait();}catch(InterruptedException e){};
  }

  void metstructsCondVarSignal() {
    notifyAll();
  }

  void metstructsCondVarWait() {
    try{wait();}catch(InterruptedException e){};
  }

  public ExecConditionChecker(Database xdb, Executive xexec,
			      ExecHasThreadedComponents xparent) {
    checkerThread = new ExecConditionCheckerThread(this,null);
    dbThread = new ExecDBMonitorThread(this, null);
    db = xdb;
    exec = xexec;
    initConditionChecker();
    checkerThread.start();
    dbThread.start();
  }

  void initConditionChecker() {
    conditionSetChanged = false;
    checkerDBChanged = false;
    savedWakeupStruct = null;
  }

  public ExecConditionCheckerStruct addConditionCheck(
				ListOfConditions  condList, 
				ListOfVariableBindings bindingList, 
				boolean checkSatisfied, 
				boolean failOnWakeup,
				Symbol id,
				boolean lock) {
    //Verify.beginAtomic(); 
    if (debug)
      System.out.println(Thread.currentThread()+"ExecConditionChecker: addConditionCheck\n");
    ExecConditionCheckerStruct condCheck = new ExecConditionCheckerStruct(condList, bindingList, checkSatisfied, failOnWakeup, id);
    if (lock) {
      if (debug)
      System.out.println(Thread.currentThread()+"ExecConditionChecker: addConditionCheck, getting mutex\n");
      //Verify.endAtomic();    
      synchronized(this) {
        //Verify.beginAtomic();    
	if (debug)
	  System.out.println(Thread.currentThread()+"ExecConditionChecker: addConditionCheck, got mutex\n");
	conditionSet.push(condCheck);
	conditionSetChanged = true;
	recheckCondVarSignal();
	if (debug)
	  System.out.println(Thread.currentThread()+"ExecConditionChecker: addConditionCheck, freeing mutex\n");
        //Verify.endAtomic();    
      }
    }
    else {
      if (debug)
	System.out.println(Thread.currentThread()+"ExecConditionChecker: addConditionCheck, got mutex\n");
      conditionSet.push(condCheck);
      conditionSetChanged = true;
      recheckCondVarSignal();
      //Verify.endAtomic();    
    }
    return condCheck;
  }
  
  void removeConditionCheck(Symbol id, boolean lock) {
    //Verify.beginAtomic();    
    if (lock) {
      //Verify.endAtomic();    
      synchronized(this) {
	//Verify.beginAtomic();    
	conditionSet.deleteConditionCheck(id);
	// Don't need to signal, since we're not adding anything new to check.
      }
    }
    else {
      conditionSet.deleteConditionCheck(id);
    }
    //Verify.endAtomic();    
  }
  
  void removeAllConditionChecks(boolean lock) {
    //Verify.beginAtomic();    
    if (lock) {
      //Verify.endAtomic();    
      synchronized (this) {
        //Verify.beginAtomic();    
	conditionSet.deleteListElements();
	// Don't need to signal, since we're not adding anything new to check.
      }
    }
    else {
      conditionSet.deleteListElements();
    }
    //Verify.endAtomic();    
  }
  
  void waitForConditions() {
    //Verify.beginAtomic();
    ExpListOfExecCCheckerStrs metConditions = new ExpListOfExecCCheckerStrs();
    if (debug)
      System.out.println(Thread.currentThread()+ "ExecConditionChecker: waiting for conditions\n");
    for (;;) {
      if (debug)
	System.out.println(Thread.currentThread()+ "ExecConditionChecker: locking mutex\n");
      //Verify.endAtomic();
      synchronized (this) {
        //Verify.beginAtomic();
	if (debug)
	  System.out.println(Thread.currentThread()+ "ExecConditionChecker: got mutex\n");
	while (!(conditionSetChanged || checkerDBChanged)) {
	  if (debug)
	    System.out.println(Thread.currentThread()+ "ExecConditionChecker: waiting for condVar: either conditions or database must change\n");
	  // TODO: pthread_cleanup_push a mutexUnlock
          //Verify.endAtomic();
	  recheckCondVarWait();
          //Verify.beginAtomic();
	}
	// TODO: need to keep db from doing another assert before we reach here?
	conditionSetChanged = false; 
	checkerDBChanged = false;

	if (debug)
	  System.out.println(Thread.currentThread()+ "ExecConditionChecker: got condVar signal\n");
	// Check all the conditions.  Do this while holding the exec from
	// doing anything by itself.
	// Construct a list of conditions that have been violated or satisfied.
	if (debug)
	  System.out.println(Thread.currentThread()+ "ExecConditionChecker: finding met conditions\n");
        //Verify.endAtomic();
	findMetConditions(metConditions, db);
        //Verify.beginAtomic();
	if (debug)
	  System.out.println(Thread.currentThread()+ "ExecConditionChecker: after finding met conditions\n");
	// Signal the Exec condVar if there are any.
      } 
      if (metConditions.empty()) {
	if (debug)
	  System.out.println(Thread.currentThread()+ "ExecConditionChecker: no met conditions\n");
      } else {
	if (debug)
	  System.out.println(Thread.currentThread()+ "ExecConditionChecker: found some met conditions, processing\n");
        //Verify.endAtomic();
	processCCStructs(metConditions);
        //Verify.beginAtomic();
	if (debug)
	  System.out.println(Thread.currentThread()+ "ExecConditionChecker: after processing met conditions\n");
	if (debug)
	  System.out.println(Thread.currentThread()+"ExecConditionChecker: after signalling exec\n");
      }
    }
    // This will never be reached.
  }

  void findMetConditions(ExpListOfExecCCheckerStrs metstructs, 
			 Database dbase) {
    //Verify.beginAtomic(); 
    Iterator iter = conditionSet.getIterator();
    while(iter.hasNext()) {
      if (debug) {
        System.out.println(Thread.currentThread()+"Checking checkerStruct ");
      }
      ExecConditionCheckerStruct exec_cc = 
	(ExecConditionCheckerStruct)iter.next();
      if (exec_cc != null) {
	ExecConditionCheckerStruct ccstruct = exec_cc;

	ListOfVariableBindings bindings = ccstruct.getBindingList_nonconst();
	int numbindings = bindings.length();
	ListOfConditions conds = ccstruct.getConditionList();
	if (debug)
	  System.out.println(Thread.currentThread()+"Before DBQueryList: ");
	//Verify.endAtomic(); 
	boolean queryResult = exec.checkConditionList(conds,bindings,false);
	//Verify.beginAtomic(); 
	if (debug)
	  System.out.println(Thread.currentThread()+"After DBQueryList, result=");
	boolean checkSatisfied = ccstruct.getCheckSatisfied();
	// If it's satisfied and checking for satisfaction,
	// or it's unsatisfied and checking for failure,
	// return.
	if ((queryResult && checkSatisfied) || 
	    (!queryResult && !checkSatisfied)) {
	  // Pop this and push it on the wakeup list.
	  // Delete the cell, but not the contents.
	  iter.remove();
	  // Put it on the list of metStructs.
	  metstructs.pushEnd(ccstruct);
      } else {
	// Reset the binding list.
	if (debug)
	  System.out.println(Thread.currentThread()+"Before trimming back, bindings=");
	for (int nb = bindings.length(); nb > numbindings; nb--) {
	  VariableBinding binding_to_delete = bindings.pop();
	  System.out.println(Thread.currentThread()+"Deleting binding ");
	  binding_to_delete = null;
	  System.out.println(Thread.currentThread()+"Remaining bindings: ");
	}
	if (debug)
	  System.out.println(Thread.currentThread()+"After clearCells, bindings=");
      }
      } else {
      }
    }
    //Verify.endAtomic(); 
  }

  public void deleteSavedWakeup() {
    savedWakeupStruct = null;
  }

  void processCCStructs(ExpListOfExecCCheckerStrs ccstructs) {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"ExecConditionChecker: locking exec mutex\n");
    //Verify.endAtomic();
    synchronized(exec) {
      //Verify.beginAtomic();
      if (debug)
	System.out.println(Thread.currentThread()+"ExecConditionChecker: got exec mutex\n");
      // Find the highest element.
      int minDepth = -1;
      ExecConditionCheckerStruct minStruct = null;
      // iterate over wakeup times.
      Iterator iter = ccstructs.getIterator();
      while(iter.hasNext()) {
	ExecConditionCheckerStruct ccstruct_p = (ExecConditionCheckerStruct)iter.next();	
	if (ccstruct_p != null) {
	  int ccstructDepth = exec.contextDepth(ccstruct_p.getID());
	  // Save this one if:
	  //  - it's in the context stack, AND
	  //  - - it's higher than the current min OR
	  //  - - it's equal and it's a success wakeup (preferred over fail).
	  if (ccstructDepth >= 0 && (minDepth < 0 || 
				     ccstructDepth > minDepth ||
				     (ccstructDepth == minDepth && 
				      minStruct.getFailOnWakeup() &&
				      !ccstruct_p.getFailOnWakeup()))) {
	    // Save this one.
	    minDepth = ccstructDepth;
	    minStruct = ccstruct_p;
	  } else {
	    ccstruct_p = null;
	  }
	}
	iter.remove();
      }
      
      if (debug)
	System.out.println(Thread.currentThread()+"ExecConditionChecker: locking conditionChecker mutex\n");
      //Verify.endAtomic();
      synchronized(this) {
        //Verify.beginAtomic();
	if (debug)
	  System.out.println(Thread.currentThread()+"ExecConditionChecker: got conditionChecker mutex");
	// Now if we found a min, save it.
	if (minStruct != null) {
	  if (debug)
	    System.out.println(Thread.currentThread()+"minstruck not null");
	  if (savedWakeupStruct != null) {
	    if (debug)
	      System.out.println(Thread.currentThread()+"savedWakeUpStruct not null");
	    // Check against former context depth and replace or discard.
	    int oldDepth = exec.contextDepth(savedWakeupStruct.getID());
	    if (oldDepth < 0 || oldDepth <= minDepth) {
	      if (debug) 
		System.out.println(Thread.currentThread()+"replace savedWakeupStruct");
	      // replace it.
	      if (debug)
		System.out.println(Thread.currentThread()+"ConditionChecker: replacing saved wakeup ");
	      savedWakeupStruct = minStruct;
	      // signal the exec
	      if (minStruct.getFailOnWakeup()) {
		// Then we set the exec signal to ConditionFailure before signalling.
		if (debug)
		  System.out.println(Thread.currentThread()+"ConditionChecker: Reporting condition failure\n");
		exec.reportExecutionSignal(ExecutionSignal.ExecSignal_ConditionFailure,false); // mutex already locked
		deleteSavedWakeup();
	      }
	      else {
		if (debug)
		  System.out.println(Thread.currentThread()+"getFailOnWakeup = false");
	      }
	      exec.notifyCondVarSignal();
	      //
	    } else {
	      if (debug) 
		System.out.println(Thread.currentThread()+"delete minStruct");
	      // forget about it and leave the old one.
	      minStruct = null;
	    }
	  } else {
	    if (debug)
	      System.out.println(Thread.currentThread()+"savedWakeUpStruct null");
	    // No saved wakeup, so save this one.
	    if (debug)
	      System.out.println(Thread.currentThread()+"ConditionChecker: replacing null saved wakeup with ");
	    savedWakeupStruct = minStruct;
	    if (minStruct.getFailOnWakeup()) {
	      // Then we set the exec signal to ConditionFailure before signalling.
	      if (debug)
		System.out.println(Thread.currentThread()+"ConditionChecker: Reporting condition failure");
	      exec.reportExecutionSignal(ExecutionSignal.ExecSignal_ConditionFailure,false); // mutex already locked
	      deleteSavedWakeup();
	    }
	    exec.notifyCondVarSignal();
	  }
	}
	else {
	  if (debug)
	    System.out.println(Thread.currentThread()+"minStruct  null");
	}
	if (debug)
	  System.out.println(Thread.currentThread()+"ExecConditionChecker: unlocking conditionChecker mutex\n");
      }      
      if (debug)
	System.out.println(Thread.currentThread()+"ExecConditionChecker: unlocking exec mutex\n");
    }
    //Verify.endAtomic();
  }

  void internalConditionCheckerLoop() {
    if (debug) 
      System.out.println(Thread.currentThread()+"ConditionChecker thread starting\n");
    waitForConditions();
  }
  
  void internalDBMonitorLoop() {
    //Verify.beginAtomic();
    if (debug) 
      System.out.println(Thread.currentThread()+ "dbMonitor thread starting\n");
    //Verify.endAtomic();
    synchronized (db) {
      //Verify.beginAtomic();
      for (;;) {
	// Wait for the DB cond var to be signaled.
	if (debug)
	  System.out.println(Thread.currentThread()+ "dbMonitor waiting\n");
	while (!db.isDBChanged()) {
          //Verify.endAtomic();
	  db.notifyCondVarWait();
          //Verify.beginAtomic();
	}
	db.resetDBChanged();
	if (debug)
	  System.out.println(Thread.currentThread()+ "dbMonitor saw change\n");
        //Verify.endAtomic();
	synchronized (this) {
          //Verify.beginAtomic();
	  checkerDBChanged = true;
	  // Signal the condition checker cond var.
	  if (debug)
	    System.out.println(Thread.currentThread()+ "dbMonitor signaling\n");
	  recheckCondVarSignal();
	}
      }
    }
  }
}
