package rover;
//import gov.nasa.jpf.jvm.Verify;

class ExecTimer extends Thread {

  class timespec {
    long tv_sec;
    long tv_nsec;
  }

  static final boolean debug = Main.print;

  ExpandableListOfExecWakeupTimes timeQueue = new ExpandableListOfExecWakeupTimes();
  boolean timeQueueChanged;
  static int TimerIndex; // = 0
  static int TimeQueueIndex; // = 1
  static int WakeupIndex; // = 2

  TimePoint  zeroTimePoint;
  timespec zeroTime = new timespec();

  CommunicationChannelQueue pastWakeupQueue = new CommunicationChannelQueue();

  void initExecTimer() {
    timeQueueChanged = false;
  }
  
  void initZeroTime() {
    long sec;
    int usec;
    sec = zeroTimePoint.numericvalSecs();
    usec = zeroTimePoint.numericvalmSecs();
    zeroTime.tv_sec = sec;
    zeroTime.tv_nsec = usec * 1000;
  }

  public ExecTimer() {    
    //WV currentTimeMillis returns mSecs since 1970 etc.
    zeroTimePoint = new TimePoint((long) (System.currentTimeMillis()/1000));
    initExecTimer();
    initZeroTime();
  }

  public ExecTimer(TimePoint zero_tp) {    
    //WV currentTimeMillis returns mSecs since 1970 etc.
    zeroTimePoint = new TimePoint(zero_tp);
    initExecTimer();
    initZeroTime();
  }


  //internalTimerLoop();
  public void run() {
    if (debug)
      System.out.println(Thread.currentThread()+"ExecTimer: internalTimerLoop starting");
    waitForNextTimePoint();
  }

  void wakeupCondVarWait() {
    try{wait();}catch(InterruptedException e){};
  }

  void wakeupCondVarSignal() {
    notifyAll();
  }

  void queueCondVarSignal() {
    notifyAll();
  }

  void queueCondVarWait() {
    try{wait();}catch(InterruptedException e){};
  }

  boolean queueCondVarTimedWait(timespec to) {
    long wait_time = to.tv_sec*1000+to.tv_nsec/1000000;
    long milliseconds = System.currentTimeMillis();    
    long new_wait_time = wait_time - milliseconds;
    if (debug)
      System.out.println(Thread.currentThread()+"Wait for " + new_wait_time);
    if (new_wait_time > 0) 
      try{wait(new_wait_time);}catch(InterruptedException e){};
    if (debug)
      System.out.println(Thread.currentThread()+"Done Waiting");
    if (System.currentTimeMillis() > wait_time) 
      return true;
    else
      return false; // no timeout
  }

  void timerCondVarWait() {
    try{wait();}catch(InterruptedException e){};
  }

  void timerCondVarSignal() {
    notifyAll();
  }

  ExecWakeupTime waitForNextTimePoint() {
    synchronized(this) {
      ////Verify.beginAtomic();
      for (;;) {
	//Verify.beginAtomic();
	if (timeQueue.empty()) {
	// Wait forever, but stop when the condition variable is signaled.
	  while (timeQueueChanged == false) {
	    if (debug)
	      System.out.println(Thread.currentThread()+"ExecTimer: Waiting for timeQueue to change");
            //Verify.endAtomic();
	    queueCondVarWait();
            //Verify.beginAtomic();
	  }
	  if (debug)
	    System.out.println(Thread.currentThread()+"ExecTimer: timeQueue changed");
	  timeQueueChanged = false;
	} else {
	  // Wait until we reach the first time point or when the condition
	  // variable is signaled.
	  ExecWakeupTime wakeup = timeQueue.head();
	  if (debug && Main.print)
	    System.out.println(Thread.currentThread()+"ExecTimer: wakeup = " + wakeup);
	  if ((wakeup != null) && wakeup.getTimePoint().isPlusInfinity()) {
	    // Then we're waiting until infinity.  Just pop this
	    // and loop, which will wait forever anyway.
	    timeQueue.pop();
	    //delete wakeup;
	    wakeup = null;
            //Verify.endAtomic();
	    continue;
	  }
	  timespec to = new timespec();
	  TimePoint curTime = getCurrentTime();
	  if (debug)
	    System.out.println(Thread.currentThread()+"ExecTimer: current time: " + curTime);
	  if (curTime.ge(wakeup.getTimePoint())) {
	    // Then signal right away without waiting.
	    if (debug)
	      System.out.println(Thread.currentThread()+"ExecTimer: Not waiting -- popping next element");
            //Verify.endAtomic();
	    popPastWakeups(curTime, wakeup.getTimePoint());
            //Verify.beginAtomic();
	    timerCondVarSignal();
	    if (debug) 
	      System.out.println(Thread.currentThread()+"ExecTimer: after condVarSignal");
	  } else {
	    // Wait, then signal if we hit a timeout before the queue changes.
	    timePointToTimespec(wakeup.getTimePoint(), to);
	    while (timeQueueChanged == false) {
	      if (debug)
		System.out.println(Thread.currentThread()+"ExecTimer: Waiting for timeQueue to change or time to reach " + wakeup.getTimePoint());
	      // TODO: pthread_cleanup_push a mutexUnlock
	      //Verify.endAtomic();
	      boolean timeout = queueCondVarTimedWait(to);
	      //Verify.beginAtomic();
	      //WV if (err == ETIMEDOUT) 
	      if (timeout) {
		/* timeout, do something */
		// Pop the first element (already saved in wakeup).
		if (debug)
		  System.out.println(Thread.currentThread()+"ExecTimer: Hit wakeup for " 
				     + wakeup.getID() +
				     " at time " + 
				     wakeup.getTimePoint());
		// pop all the time points reached before signalling.
		//Verify.endAtomic();
		popPastWakeups(curTime, wakeup.getTimePoint());
		//Verify.beginAtomic();
		if (debug)
		  System.out.println(Thread.currentThread()+"ExecTimer: time reached");
		timerCondVarSignal();
		if (debug)
		  System.out.println(Thread.currentThread()+"ExecTimer: after condVarSignal");
		break;
	      } else if (timeQueueChanged) {
		if (debug)
		  System.out.println(Thread.currentThread()+"ExecTimer: timeQueue changed");
		timeQueueChanged = false;
		break;
	      }
	    }
	  }
	  curTime = null;
	}
	//Verify.endAtomic();
      }
      // We should never get here.
      // TODO: unlock the mutex in the thread cleanup code.
    }
  }

  void popPastWakeups(TimePoint now, TimePoint firstTime) {
    //Verify.beginAtomic(); 
    // First wait until the wakeup list is cleared by the process that's
    // reading it.
    // TODO: VERIFY that the wait for the wakeup list to be cleared is
    // short.
    if (!pastWakeupQueue.EmptyCommunicationChannel()) {
      if (debug)
	System.out.println(Thread.currentThread()+"ExecTimer: Waiting for wakeup condVar");
      // TODO: pthread_cleanup_push a mutexUnlock
      //Verify.endAtomic(); 
      wakeupCondVarWait();
      //Verify.beginAtomic(); 
    }
    // Now add all the elements that are <= than the current time or the
    // first element of the list (because of < 1 sec roundoff when the
    // thread starts, we allow either).
    while (!timeQueue.empty()) {
      ExecWakeupTime first = timeQueue.head();
      if ((first!=null) && (now!=null) && (firstTime!=null) && 
	  (now.gt(first.getTimePoint()) || 
      	   firstTime.gt(first.getTimePoint()))) {
	// Pop this and push it on the wakeup list.
	ExecWakeupTime nextWakeup = timeQueue.pop();
	if (debug)
	  System.out.println(Thread.currentThread()+"ExecTimer: pushing TimePoint " + 
			     nextWakeup.getTimePoint() +
			     " onto wakeup list");
	pastWakeupQueue.AddToCommunicationChannel(nextWakeup);
      } else {
	break;
      }
    }
    // When we get here, we've either emptied the timeQueue or have elements
    // that need waiting, so return.
    //Verify.endAtomic();
    return;
  }

  class timeval {
    long tv_sec;
    long tv_usec;
  }

  TimePoint getCurrentTime() {
    //  TimePoint tp((unsigned long) (time(0)));
    //    struct timeval now;
    //gettimeofday(&now, 0);
    timeval now = new timeval();
    long milliseconds = System.currentTimeMillis();
    now.tv_usec = (milliseconds % 1000) * 1000;
    now.tv_sec  = milliseconds/1000;
    now.tv_sec -= zeroTime.tv_sec;
    // Changed to make sure unsigneds never try to go negative.
    if (now.tv_usec < zeroTime.tv_nsec / 1000) {
      now.tv_usec += 1000000 - zeroTime.tv_nsec / 1000;
      now.tv_sec--;
    } else {
      now.tv_usec -= zeroTime.tv_nsec / 1000;
    }
    return (Factory.newTimePoint(now.tv_sec, (int)now.tv_usec / 1000));
  }

  void timePointToTimespec(TimePoint tp, timespec to) {
    long secs;
    int msecs;
    secs = tp.numericvalSecs();
    msecs = tp.numericvalmSecs();;

    long zero_secs;
    int zero_msecs;
    zero_secs = zeroTimePoint.numericvalSecs();
    zero_msecs = zeroTimePoint.numericvalmSecs();

    to.tv_sec = zero_secs + secs;
    int total_msecs = zero_msecs + msecs;
    if (total_msecs >= 1000) {
      to.tv_sec += total_msecs / 1000;
      total_msecs %= 1000;
    }
    to.tv_nsec = total_msecs * 1000000;
  }

  void removeTime(Symbol id) {
    synchronized(this) {
      timeQueue.deleteWakeup(id);
      if (debug && Main.print)
	System.out.println(Thread.currentThread()+"ExecTimer: timeQueue after removeTime: " + 
			 timeQueue);
      timeQueueChanged = true;
      queueCondVarSignal();
    }
  }

  void removeAllTimes() {
    synchronized(this) {
      timeQueue.deleteListElements();
      if (debug && Main.print)
	System.out.println(Thread.currentThread()+"ExecTimer: timeQueue after removeTime: " +
			   timeQueue);
      timeQueueChanged = true;
      queueCondVarSignal();
    }
  }

  ExecWakeupTime addTime(TimePoint tp, 
			 Symbol id, 
			 boolean failOnWakeup,
			 Executive exec) {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"ExecTimer (from Executive): addTime");
    if (!tp.isnumeric()) {
      if (debug)
	System.out.println(Thread.currentThread()+"ExecTimer: Trying to add non-numeric time " + tp);
      //Verify.endAtomic();
      return null;
    }
    ExecWakeupTime wakeup = new ExecWakeupTime(tp, id, failOnWakeup, exec);
    //Verify.endAtomic();
    synchronized(this) {
      //Verify.beginAtomic();
      if (debug && Main.print)
	System.out.println(Thread.currentThread()+"ExecTimer (from Executive): addTime ready to do Insert of " + wakeup + " in queue = " + timeQueue);
      timeQueue.sortedInsert(wakeup);
      if (debug && Main.print)
	System.out.println(Thread.currentThread()+"ExecTimer: timeQueue after addTime: " + timeQueue);
      timeQueueChanged = true;
      queueCondVarSignal();
      //Verify.endAtomic();
    }
    return wakeup;
  }

  CommunicationChannelQueue getPastWakeups() {
    return pastWakeupQueue; 
  }

}
