package rover;

//import gov.nasa.jpf.jvm.Verify;

import java.util.Iterator;

class ExecTimerWaiter extends Thread // ExecThread
{

	final static boolean debug = Main.print;

	public static class ExecTimerWaiterOverrideFn implements
			CommunicationChannelSingleVariable.OverrideFnType {
		public boolean override(Object newObject, Object oldObject) {
			ExecWakeupTime newWakeup;
			ExecWakeupTime oldWakeup;

			newWakeup = (ExecWakeupTime) newObject;
			oldWakeup = (ExecWakeupTime) oldObject;

			int newWakeupDepth = newWakeup.getExecutive().contextDepth(
					newWakeup.getID());
			int oldWakeupDepth = oldWakeup.getExecutive().contextDepth(
					oldWakeup.getID());

			if (newWakeupDepth >= 0
					&& (oldWakeupDepth < 0 || newWakeupDepth > oldWakeupDepth || (newWakeupDepth == oldWakeupDepth
							&& oldWakeup.getFailOnWakeup() && !(newWakeup
							.getFailOnWakeup()))))
				return true;
			else
				return false;
		}
	}

	final private Executive exec;

	final private ExecTimer timer;

	private CommunicationChannelSingleVariable savedWakeupTime;

	public ExecTimerWaiter(Executive xexec, ExecTimer xtimer) {
		this.exec = xexec;
		this.timer = xtimer;
		this.savedWakeupTime = new CommunicationChannelSingleVariable(
				new ExecTimerWaiterOverrideFn(), true, "ExecTimerWaiter");
		start();
	}

	public void run() {
		runTimerWaitThreadInternal();
	}

	public ExecWakeupTime getSavedWakeup() {
		return (ExecWakeupTime) savedWakeupTime
				.PeekCommunicationChannelReader();
	}

	public void clearSavedWakeup(boolean writer) {
		if (writer)
			savedWakeupTime.ClearCommunicationChannelWriter();
		else
			savedWakeupTime.ClearCommunicationChannelReader();
	}

	public void runTimerWaitThreadInternal() {
		ExpandableListOfExecWakeupTimes wakeups;

		//Verify.beginAtomic();

		if (debug)
			System.out.println(Thread.currentThread()+"TimerWaiter: Timer wait thread starting\n");
		for (;;) {
			//Verify.endAtomic();
			synchronized (timer) {
				//Verify.beginAtomic();
				if (debug)
					System.out
							.println("TimerWaitter: timer_wait: Waiting for timer condVar\n");
				// TODO: make sure condvar wait is terminated for right reason!
				while (timer.getPastWakeups().EmptyCommunicationChannel()) {
					//Verify.endAtomic();
					timer.timerCondVarWait();
					//Verify.beginAtomic();
				}

				if (debug)
					System.out
							.println("TimerWaiter: timer_wait: got condVar signal\n");
				// Copy the wakeup list elements so that when the past wakeups are
				// cleared, this doesn't bomb.
				// TODO: Verify that this is necessary.
				CommunicationChannelQueue pastWakeups = timer.getPastWakeups();
				// Copy out the past wakeups.
				wakeups = new ExpandableListOfExecWakeupTimes();
				while (!pastWakeups.EmptyCommunicationChannel()) {
					ExecWakeupTime newWakeup = (ExecWakeupTime) pastWakeups
							.ReadCommunicationChannel();
					wakeups.pushEnd(newWakeup);
				}
				if (debug)
					System.out
							.println("TimerWaiter: before clearPastWakeups, wakeup list length="
									+ wakeups.length());
				if (debug)
					System.out
							.println("TimerWaiter: after clearPastWakeups, wakeup list length="
									+ wakeups.length());
				// Now we don't care if the timer changes while we're processing
				// these structs.
				//Verify.endAtomic();
				processWakeupTimes(wakeups);
				//Verify.beginAtomic();
				// The wakeups list should be empty by the time we reach here,
				// so the following statement may not be necessary.
				wakeups.deleteListElements();
				if (debug)
					System.out
							.println("timer_wait: wakeups empty, signaling wakeup condvar");
				timer.wakeupCondVarSignal();
			}
		}
	}

	void processWakeupTimes(ExpandableListOfExecWakeupTimes wakeups) {
		// Iterate over wakeup structs, choose the one that affects the
		// highest element of the context stack, then hold that and signal
		// the exec.
		//Verify.beginAtomic();
		if (debug)
			System.out.println(Thread.currentThread()+"TimerWaiter: waiting for exec mutex");
		//Verify.endAtomic();
		synchronized (exec) {
			//Verify.beginAtomic();
			if (debug)
				System.out.println(Thread.currentThread()+"TimerWaiter: got exec mutex");
			// Lock the exec mutex while examining the context stack.
			// Find the highest element.
			int minDepth = -1;
			ExecWakeupTime minWakeup = null;
			// iterate over wakeup times.
			Iterator iter = wakeups.getIterator();
			ExecWakeupTime wakeup_tp;

			if (debug)
				System.out.println(Thread.currentThread()+"TimerWaiter: wakeup list length="
						+ wakeups.length());
			while (iter.hasNext()) {
				wakeup_tp = (ExecWakeupTime) iter.next();
				if (wakeup_tp != null) {
					int wakeupDepth = exec.contextDepth(wakeup_tp.getID());
					if (debug)
						System.out.println(Thread.currentThread()+"TimerWaiter: wakeupDepth="
								+ wakeupDepth);
					// Save this one if:
					//  - it's in the context stack, AND
					//  - - it's higher than the current min OR
					//  - - it's equal and it's a success wakeup (preferred over fail).
					if (wakeupDepth >= 0
							&& (minDepth < 0 || wakeupDepth > minDepth || (wakeupDepth == minDepth
									&& minWakeup.getFailOnWakeup() && !(wakeup_tp
									.getFailOnWakeup())))) {
						// Save this one.
						minDepth = wakeupDepth;
						if (minWakeup != null)
							//WV	      delete minWakeup;
							minWakeup = null;
						minWakeup = wakeup_tp;
					} else {
						//WV delete *wakeup_tp;
						wakeup_tp = null;
					}
				}
				// Delete the cell but not the contents.
				iter.remove();
			}
			if (debug)
				System.out.println(Thread.currentThread()+"TimerWaiter: minWakeup found " + minWakeup);
			// Now if we found a min, save it.
			if (minWakeup != null) {
				if (minWakeup == savedWakeupTime
						.AddToCommunicationChannel(minWakeup)) {
					// Replaced previous value.
					if (minWakeup.getFailOnWakeup()) {
						// Then we set the exec signal to Timeout before signalling.
						if (debug)
							System.out
									.println("TimerWaiter: Reporting action failure");
						exec.reportExecutionSignal(
								ExecutionSignal.ExecSignal_Timeout, false); // mutex already locked
						// Delete the savedWakeup, since the signal will do the trick.
						// TODO: CHECK THIS with verification: This may be a problem.
						clearSavedWakeup(true);
					}
					if (debug)
						System.out
								.println("TimerWaiter: signalling exec condvar");
					exec.notifyCondVarSignal();
				}
			}
			//Verify.endAtomic();
		}
	}
}
