package rover;
import java.util.Stack;
import java.util.Iterator;

class ExecutionContext {

  final static boolean debug = false;

  boolean inheritedContinueOnFailure;
  static boolean defaultInheritedContinueOnFailure = false;
  Node plan;
  Stack object_stack = new Stack();

  public Iterator getIterator() {
    return object_stack.iterator();
  }

  public void printContext() {
    for (Iterator i = object_stack.iterator(); i.hasNext();) {    
      ExecutionContextObject contextObj = (ExecutionContextObject)i.next();
      System.out.println("context = " + contextObj.get_node());
    }
  }

  public  ExecutionContext(Node xplan) {
    ListOfVariableBindings bindings = new  ListOfVariableBindings();
    plan = xplan;
    ExecutionContextObject contextObj = add_node(plan, bindings); 
    //System.out.println("CHECK THIS TOP: " + top_object());
    //System.out.println("CHECK THIS: " + top_bindings());
    inheritedContinueOnFailure = defaultInheritedContinueOnFailure;
    contextObj.setInheritedContinueOnFailure(inheritedContinueOnFailure);
  }

  public Node get_currentNode() {
    ExecutionContextObject top_object = (ExecutionContextObject)object_stack.peek();
    if (top_object.get_object_type() == ExecutionContextObject.NodeObject) {
      return top_object.get_node();
    } else {
      System.out.println("Top of context stack not a Node");
      return null;
    }
  }
  
  public int contextDepth(Symbol id) {
    if (id == null)
      return -1;
    for (Iterator i = object_stack.iterator(); i.hasNext();) {
      int depth = 0;
      ExecutionContextObject contextObj = (ExecutionContextObject)i.next();
      if (contextObj.get_object_type() == ExecutionContextObject.NodeObject) {
	Node nodeObj = contextObj.get_node();
	if (nodeObj.get_id().equal(id))
	  return depth;
      }
      depth++;
    }
    return -1;
  }


  public boolean  emptyExecutionContext() { 
    return (object_stack.isEmpty()); 
  }

  public ExecutionContextObject top_object() {
    if (object_stack.isEmpty())
      return null;
    else
      return (ExecutionContextObject)object_stack.peek();
  }

  public ListOfVariableBindings top_bindings() {
    ExecutionContextObject topObj = top_object();
    //System.out.println("top_object = " + topObj);
    //System.out.println("top_object.bindings = " + topObj.get_bindings());
    return topObj.get_bindings();
  }

  public Node top_node() {
    ExecutionContextObject topObj = top_object();
    return topObj.get_node();
  }

  public Option top_option() {
    ExecutionContextObject topObj = top_object();
    return topObj.get_option();
  }

  public ExecutionContextObject pop_object() {
    ExecutionContextObject topObj = (ExecutionContextObject)object_stack.pop();
    calculateInheritedContinueOnFailure();
    return topObj;
  }

  public Node pop_node() {
    // cout << "Popping node\n";
    ExecutionContextObject topObj = (ExecutionContextObject)object_stack.pop();
    Node top_node = topObj.get_node();
    //WV delete topObj;
    topObj = null;
    calculateInheritedContinueOnFailure();
    return top_node;
  }

  public Option pop_option() {
    // cout << "Popping option\n";
    ExecutionContextObject topObj = (ExecutionContextObject)object_stack.pop();
    Option top_option = topObj.get_option();
    //WV delete topObj;
    topObj = null;
    calculateInheritedContinueOnFailure();
    return top_option;
  }

  public boolean getInheritedContinueOnFailure() {
    return inheritedContinueOnFailure;
  }

  public boolean saveInheritedContinueOnFailure() {
  // Calculate what the inheritedContinueOnFailure will be for a node
  // not yet pushed onto the context stack.  If the top of the stack
  // is a node, then combine its default_contine_on_failure with 
  // its inheritedContinueOnFailure to
  // produce the new inheritedContinueOnFailure.
    ExecutionContextObject top_obj = top_object();
    if (top_obj == null)
      inheritedContinueOnFailure = defaultInheritedContinueOnFailure;
    else if (top_obj.get_object_type() == 
	     ExecutionContextObject.NodeObject) {
      inheritedContinueOnFailure = 
	top_node().get_default_continue_on_failure().boolval(inheritedContinueOnFailure);
    }
    // Else (for options/alternate plans) just retain 
    // the current continueonfailure.
    return inheritedContinueOnFailure;
  }

  public boolean calculateInheritedContinueOnFailure() {
    // Calculate what the inheritedContinueOnFailure is for an
    // existing context stack.  This is just a retrieval of the top
    // object and its inheritedContinueOnFailure.
    ExecutionContextObject top_obj = top_object();
    if (top_obj == null)
      inheritedContinueOnFailure = defaultInheritedContinueOnFailure;
    else 
      inheritedContinueOnFailure = top_obj.getInheritedContinueOnFailure();
    return inheritedContinueOnFailure;
  }

  ExecutionContextObject add_task(Task new_task, 
				  ListOfVariableBindings bindings) {
    //    System.out.println("Adding Task: ");
    // cout << "Adding task\n";
    ExecutionContextObject contextObj = new ExecutionContextObject(new_task,
								   bindings);
    //System.out.println("Pushing contectObj = " + contextObj);
    saveInheritedContinueOnFailure();
    object_stack.push(contextObj);
    return contextObj;
  }

  ExecutionContextObject add_block(Block new_block, 
				   ListOfVariableBindings bindings) {
    //System.out.println("Adding Block: ");
    // cout << "Adding block\n";
    ExecutionContextObject contextObj = new ExecutionContextObject(new_block,
								   bindings);
    saveInheritedContinueOnFailure();
    object_stack.push(contextObj);
    return contextObj;
  }

  ExecutionContextObject add_branch(Branch new_branch, 
				    ListOfVariableBindings bindings) {
    //System.out.println("Adding Branch: ");
    // cout << "Adding branch\n";
    ExecutionContextObject contextObj = new ExecutionContextObject(new_branch,
								   bindings);
    saveInheritedContinueOnFailure();
    object_stack.push(contextObj);
    return contextObj;
  }

  public ExecutionContextObject add_node(Node node, 
					 ListOfVariableBindings bindings) {
    //System.out.println("Adding Node: ");
    // cout << "Adding node\n";
    ExecutionContextObject contextObj = null;
    switch (node.getSubtype()) {
    case Node.NodeSubtypeTask: 
      contextObj = add_task((Task) node, bindings);
      break;
    case Node.NodeSubtypeBranch: 
      contextObj = add_branch((Branch) node, bindings);
      break;
    case Node.NodeSubtypeBlock: 
      contextObj = add_block((Block) node, bindings);
      break;
    default:
      System.out.println("ExecutionContext::add_node: undefined node type");
      break;
    }
    return contextObj;
  }

  ExecutionContextObject add_option(Option new_option, 
				   ListOfVariableBindings bindings) {
    // cout << "Adding option\n";
    ExecutionContextObject contextObj = new ExecutionContextObject(new_option,
								   bindings);
    boolean savedval = saveInheritedContinueOnFailure();
    contextObj.setInheritedContinueOnFailure(savedval);
    object_stack.push(contextObj);
    return contextObj;
  }

  ExecutionContextObject add_alternate_plan(AlternatePlan new_alternate_plan, 
					    ListOfVariableBindings bindings) {
    // cout << "Adding alternate_plan\n";
    ExecutionContextObject contextObj =
      new ExecutionContextObject(new_alternate_plan, bindings);
    boolean savedval = saveInheritedContinueOnFailure();
    contextObj.setInheritedContinueOnFailure(savedval);
    object_stack.push(contextObj);
    return contextObj;
  }

  void clearExecutionContext() {
    //WV    object_stack.deleteListElements();
    object_stack = null;
    plan = null;
  }

}
