package rover;
class ExecutionContextObject {

  /* ObjecType */
  public static final int NodeObject   = 0;
  public static final int OptionObject = 1;

  int /* ObjectType */ object_type;
  /* union */
  Node current_node;
  Option current_option;
  int /* NodeType */node_type;
  int /*OptionType*/ option_type;

  int positionInNode; /* for Branches, option #; for Blocks, node # */
  boolean inheritedContinueOnFailure;
  TimePoint transitionTime;
  TimePoint startTime;

  ListOfVariableBindings bindings = new ListOfVariableBindings();

  public String toString() {
    return " " + current_node;
  }

  int /*ObjectType*/ get_object_type() { 
    return object_type; 
  }

  public ExecutionContextObject() {
  }

  void initBindings(ListOfVariableBindings binds) {
    binds.copyListElements(bindings); 
  }
  
  public ExecutionContextObject(Option option, 
				ListOfVariableBindings binds) { 
    initBindings(binds);
    object_type = OptionObject;
    current_option = option;
    option_type = OptionType.OptionType_Option;
    transitionTime = null;
    startTime = null;
  }

  public ExecutionContextObject(AlternatePlan altplan, 
				ListOfVariableBindings binds) { 
    initBindings(binds);
    object_type = OptionObject;
    current_option = altplan;
    option_type = OptionType.OptionType_AlternatePlan;
    transitionTime = null;
    startTime = null;
  }

  public ExecutionContextObject(Task task, ListOfVariableBindings binds) { 
    // cout << "Making ExecutionContextObject(Task) for Task " << task;
    initBindings(binds);
    object_type = NodeObject;
    current_node = task;
    node_type = NodeType.NodeType_Task;
    transitionTime = null;
    startTime = null;
  }

  public ExecutionContextObject(Block block, ListOfVariableBindings binds) { 
    // cout << "Making ExecutionContextObject(Block) for Block " << block;
    initBindings(binds);
    object_type = NodeObject;
    current_node = block;
    node_type = NodeType.NodeType_Block;
    positionInNode = 0;
    transitionTime = null;
    startTime = null;
  }

  public ExecutionContextObject(Branch branch, ListOfVariableBindings binds) { 
    // cout << "Making ExecutionContextObject(Branch) for Branch " << branch;
    positionInNode = -1; // No choice yet.
    initBindings(binds);
    object_type = NodeObject;
    current_node = branch;
    node_type = NodeType.NodeType_Branch;
    transitionTime = null;
    startTime = null;
  }

  public int getPositionInNode() { 
    return positionInNode; 
  }

  public int incPositionInNode() { 
    return (++positionInNode); 
  }

  public void setPositionInNode(int newval) { 
    positionInNode = newval; 
  }

  public TimePoint getTransitionTime() {
    return transitionTime; 
  }

  public void setTransitionTime(TimePoint xtime) { 
    transitionTime = xtime; 
  }

  public TimePoint getStartTime() { 
    return startTime; 
  }

  public void setStartTime(TimePoint xtime) { 
    startTime = xtime; 
  }

  public /*NodeType*/ int get_node_type() { 
    if (object_type == NodeObject)
      return (node_type);
    else
      return (NodeType.UnknownNodeType);
  }

  public /*OptionType*/ int get_option_type() { 
    if (object_type == OptionObject)
      return (option_type);
    else
      return (OptionType.UnknownOptionType);
  }

  public Node get_node() {
    if (object_type == NodeObject)
      return current_node;
    else
      return null;
  }

  public Option get_option() {
    if (object_type == OptionObject)
      return current_option;
    else
      return null;
  }

  public ListOfVariableBindings get_bindings() { 
    return bindings; 
  }
  
  public boolean getInheritedContinueOnFailure() { 
    return inheritedContinueOnFailure; 
  }
  
  public void setInheritedContinueOnFailure(boolean newval) {
    inheritedContinueOnFailure = newval; 
  }


}
