package rover;
// Translation started 3/14/2002
//import gov.nasa.jpf.jvm.Verify;
import java.util.Vector;

// The //WV comment indicates removed for model checking

class Executive
  extends java.lang.Thread {
  //WV inherits from MutexAcknowledge,ExecHasThreadedComponents
  final static int EXEC_STATUS_INVALID = -1;
  final static int EXEC_STATUS_EXECUTING = 1203;
  final static int EXEC_STATUS_FAULT = 1102;
  final static int EXEC_STATUS_INTERRUPTED = 1100;
  final static int EXEC_STATUS_DONE = 1104;
  static final boolean debug = Main.print;
  static final boolean verbose = false; // Just tells what's going on.

  /* private */
  ExecConditionChecker conditionChecker;
  /* private */
  ExecTimer timer;
  /* private */
  ExecTimerWaiter timerWaiter;
  /* private */
  ActionExecution actionThread;
  PlanWatcher watcher;

  ExecutionContext context;
  Node plan;
  ListOfAlternatePlans globalAlternatePlans;
  CommandTable commandTable;
  FunctionTable functionTable;
  Database db;

  // TODO: should the status be on the context object?
  static String execStatusVarString;
  static String execReferenceTimeVarString = "";
  static String getExecStatusVarString() {
    return execStatusVarString;
  }
  static String getExecReferenceTimeVarString() {
    return execReferenceTimeVarString;
  }

  int /*ExecutionSignal*/ pendingExecSignal; //enumerated type
  Symbol signaledNodeID;

  int currentCommandID;
  int getNewCommandID() {
    return (currentCommandID++);
  }

  String pendingPlanFile;
  ExecutionContext getContext() {
    return context;
  }

  int /*ExecStatus*/ execStatus; // enumerated type
  int numericExecutionStatus; // This uses the global status numbers.

  int /* ExecStatus */ getExecStatus() {
    return execStatus;
  }

  int getNumericExecutionStatus() {
    return numericExecutionStatus;
  }

  void setNumericExecutionStatus(int status, boolean lock) {
    if (lock) {
      synchronized (this) {
        //WV mutexLock();
        numericExecutionStatus = status;
      }
      //WV mutexUnlock();
    } else {
      numericExecutionStatus = status;
    }
  }

  public Executive() {
    super();
  }

  public Executive(
    CommandTable cmdtab,
    FunctionTable fntab,
    ExternalSystem externSys,
    ExecDisplay xdisplay,
    TelemetryInterface xtelemInterface) {
    //Verify.beginAtomic();
    initialize();
    db = new Database(); 
    conditionChecker = new ExecConditionChecker(db, this, null);
    timer = Factory.newExecTimer();

    timerWaiter = new ExecTimerWaiter(this, timer);
    watcher = new PlanWatcher(this, null);
    actionThread = new ActionExecution(this, externSys, null);

    commandTable = cmdtab;
    functionTable = fntab;
    //Verify.endAtomic();
  }

  void initialize() {
    plan = null;
    context = null;
    globalAlternatePlans = null;
    commandTable = null;
    currentCommandID = 1;
    pendingExecSignal = ExecutionSignal.ExecSignal_Okay;
    pendingPlanFile = null;
    execStatus = ExecStatus.ExecStatus_Idle;
    numericExecutionStatus = EXEC_STATUS_INVALID;
  }

  void initializeContext() {
    context = new ExecutionContext(plan);
    ListOfVariableBindings bindings = context.top_bindings();
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: execution context");
  }

  void initializePlan(Node newplan) {
    plan = newplan;
  }

  /* private */
  boolean executePlanbody() throws ExecutionException {
    //Verify.beginAtomic();
    boolean terminate = false;
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: executePlan");
    resetExecutionSignal();
    if (debug || verbose)
      System.out.println(Thread.currentThread()+"Executive: initializing context");
    initializeContext();
    try {
      while (!emptyExecutionContext()) {
        if (debug || verbose)
          System.out.println(Thread.currentThread()+"Executive: executing current node");
        Node cn = context.get_currentNode();
        //Verify.endAtomic();
        executeCurrentNode();
        //Verify.beginAtomic();
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: advancing context");
        advanceContext();
      }
      if (debug || verbose)
        System.out.println(Thread.currentThread()+"Executive: done executing plan");
    } catch (ExecutionException except) {
      //Verify.beginAtomic();
      switch (except.getException()) {
        case ExecutionException.NodeFailure :
          //set flag to failure 
          numericExecutionStatus = EXEC_STATUS_FAULT;
          if (debug || verbose)
            System.out.println(Thread.currentThread()+
              "Executive: caught node failure at top level, aborting plan");
          break;
        case ExecutionException.PlanAbort :
          //set flag to interrupted 
          numericExecutionStatus = EXEC_STATUS_INTERRUPTED;
          if (debug || verbose)
            System.out.println(Thread.currentThread()+"Executive: caught plan abort at top level, aborting plan");
          break;
        case ExecutionException.ExecTerminate :
          //set flag to interrupted 
          numericExecutionStatus = EXEC_STATUS_INTERRUPTED;
          if (debug || verbose)
            System.out.println(Thread.currentThread()+
              "Executive: caught exec terminate at top level, aborting plan");
          terminate = true;
          break;
        default :
          if (debug || verbose)
            System.out.println(Thread.currentThread()+"Executive: caught ExecutionException at top level");
      }
    } catch (Exception e) {
      if (Main.JPF) {
	  if (e instanceof RuntimeException)
	    System.out.println(Thread.currentThread()+"Runtime Exception!");
          else 
	    System.out.println(Thread.currentThread()+"Unknown Exception aborting");
	  assert(false);
      }
      if (debug)
        System.out.println(Thread.currentThread()+
          "Executive: caught unknown exception at top level, aborting plan " + e);
      if (debug && Main.print) {
        System.out.println(Thread.currentThread()+
          "Executive: caught unknown exception at top level, aborting plan " + e);
        e.printStackTrace();
      }
    }
    //Verify.endAtomic();
    conditionChecker.removeAllConditionChecks(true);
    timer.removeAllTimes();
    //Verify.beginAtomic();
    // Clear the execution context.
    if (!emptyExecutionContext()) {
      context.clearExecutionContext();
    }
    // Clear the plan.
    initializePlan(null);
    // In case anyone's waiting for the end of the plan, signal it.
    notifyCondVarSignal();

    // If we didn't get any other status, set it to DONE.
    if (numericExecutionStatus == EXEC_STATUS_EXECUTING) {
      numericExecutionStatus = EXEC_STATUS_DONE;
    }

    execStatus = ExecStatus.ExecStatus_Idle;
    //Verify.endAtomic();
    return terminate;
  }

  void executePlan(boolean lock) throws ExecutionException {
    //Verify.beginAtomic();
    boolean terminate = false;
    numericExecutionStatus = EXEC_STATUS_EXECUTING;

    if (lock) {
      //Verify.endAtomic();
      synchronized (this) {
        terminate = executePlanbody();
      }
    } else {
      //Verify.endAtomic();
      terminate = executePlanbody();
    }
    // If terminating exec, throw the signal instead of returning.
    if (terminate)
      throw new ExecutionException(ExecutionException.ExecTerminate);
  }

  void advanceContext() {
    // Advance context to next step in plan.
    if (context == null)
      return;
    // Pop off the top of the stack.
    ExecutionContextObject topObj = context.top_object();
    while (topObj != null) {
      if (topObj.get_object_type() == ExecutionContextObject.NodeObject
        && topObj.get_node().getSubtype() == Node.NodeSubtypeBlock) {
        // Top of stack is a block; advance the block's pointer.
        if (debug)
          System.out.println(Thread.currentThread()+"Found block on context stack\n");
        ListOfNodes nodelist = ((Block) (topObj.get_node())).getNodeList();
        topObj.incPositionInNode();
        if ((nodelist != null) && topObj.getPositionInNode() < nodelist.length())
          // Stop popping the context when you hit a block that's not finished.
          break;
      }
      // Otherwise pop the stack and delete it.
      if (debug)
        System.out.println(Thread.currentThread()+"Popping top of context stack " + topObj);
      context.pop_object();
      //WV delete topObj;
      topObj = context.top_object();
    }
  }

  void assertCommandStatus(int /*ExecCmdIDType*/ cmdID, int status) {
    // TODO: change the exec so it can assert non-termination status without
    // screwing up the waitForActionTermination function.
    //Verify.beginAtomic();
    if (status == ActionExecution.EXEC_RETURNVALUE_SUCCESS
      || status == ActionExecution.EXEC_RETURNVALUE_FAILURE
      || status == ActionExecution.EXEC_RETURNVALUE_TERMINATE) {
      if (debug || verbose)
        System.out.println(Thread.currentThread()+"Executive: Asserting cmd status assertion ");
      //Verify.endAtomic();
      db.DBAssert();
      //Verify.beginAtomic();
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: Done asserting cmd status assertion");
    } else {
      if (debug || verbose)
        System.out.println(Thread.currentThread()+
          "Executive: Not asserting non-termination cmd status "
            + status
            + " for cmdID "
            + cmdID);
    }
    //Verify.endAtomic();
  }

  void assertTransitionTime(Symbol nodeID, double refTime) {
    assertNodeTime(nodeID, refTime, NodeTimeType.NodeTransitionTime);
  }

  void assertStartTime(Symbol nodeID, double refTime) {
    assertNodeTime(nodeID, refTime, NodeTimeType.NodeStartTime);
  }

  void assertEndTime(Symbol nodeID, double refTime) {
    assertNodeTime(nodeID, refTime, NodeTimeType.NodeEndTime);
  }

  void assertReferenceTime(Symbol nodeID, double refTime) {
    Symbol nodeIDCopy;
    if (nodeID != null)
      nodeIDCopy = new Symbol(nodeID);
    else
      nodeIDCopy = null;
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: Asserting reference time assertion ");
    //Verify.endAtomic();
    db.DBAssert();
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: Done asserting reference time assertion\n");
  }

  void assertNodeTime(
    Symbol nodeID,
    double refTime,
    int /*NodeTimeType*/ timeType) {

    switch (timeType) {
      case NodeTimeType.NodeTransitionTime :
        {
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: Asserting node time assertion ");
          //Verify.endAtomic();
          db.DBAssert();
          //Verify.beginAtomic();
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: Done asserting node time assertion");
          break;
        }
      case NodeTimeType.NodeStartTime :
        {
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: Asserting node time assertion ");
          //Verify.endAtomic();
          db.DBAssert();
          //Verify.beginAtomic();
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: Done asserting node time assertion\n");
          break;
        }
      case NodeTimeType.NodeEndTime :
        {
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: Asserting node time assertion ");
          //Verify.endAtomic();
          db.DBAssert();
          //Verify.beginAtomic();
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: Done asserting node time assertion");
          break;
        }
    }
  }

  TimePoint setAndAssertNodeTime(
    int /*NodeTimeType*/ timeType,
    Symbol nodeID,
    ExecutionContextObject currentContextObject) {
    //WV precondition this is in an atomic block!
    // Returns the current time as a double.
    TimePoint curTime = timer.getCurrentTime();
    double doubleTime = curTime.doubleval();
    switch (timeType) {
      case NodeTimeType.NodeTransitionTime :
        currentContextObject.setTransitionTime(curTime);
        assertTransitionTime(nodeID, doubleTime);
        assertReferenceTime(nodeID, doubleTime);
        break;
      case NodeTimeType.NodeStartTime :
        currentContextObject.setStartTime(curTime);
        assertStartTime(nodeID, doubleTime);
        assertReferenceTime(nodeID, doubleTime);
        break;
      case NodeTimeType.NodeEndTime :
        assertEndTime(nodeID, doubleTime);
        break;
    }
    return curTime;
  }

  static class TimePointReference {
    TimePoint tp;
  }

  void getStartTimeBounds(
    Node current_node,
    TimePoint curTime,
    TimePointReference startLB,
    TimePointReference startUB) {
    // startLB is the max of the abs start LB and the rel start LB + curTime.
    // startUB is the min of the abs start UB and the rel start UB + curTime.
    TimeCondition absStart = current_node.getAbsoluteStartBounds();
    TimePoint absUB;
    if (absStart != null)
      absUB = absStart.getEndTime();
    else
      absUB = null;
    TimeCondition relStart = current_node.getRelativeStartBounds();
    if (relStart != null) {
      TimePoint relUB = relStart.getEndTime(); // really a TimeDate*.
      TimePoint offsetUB = curTime.plus(relUB);
      if (absUB == null || offsetUB.lt(absUB))
        startUB.tp = offsetUB;
      else {
        //delete offsetUB;
        offsetUB = null;
        startUB.tp = new TimePoint(absUB);
      }
    } else {
      if (absUB != null)
        startUB.tp = new TimePoint(absUB);
      else
        startUB.tp = null;
    }
    TimePoint absLB;
    if (absStart != null)
      absLB = absStart.getStartTime();
    else
      absLB = null;
    if (relStart != null) {
      TimePoint relLB = relStart.getStartTime(); // really a TimeDate*.
      TimePoint offsetLB = curTime.plus(relLB);
      if (absLB == null || offsetLB.gt(absLB))
        startLB.tp = offsetLB;
      else {
        //delete offsetLB;
        offsetLB = null;
        startLB.tp = new TimePoint(absLB);
      }
    } else {
      if (absLB != null)
        startLB.tp = new TimePoint(absLB);
      else
        startLB.tp = null;
    }
  }

  void getEndTimeBounds(
    Node current_node,
    TimePoint curTime,
    TimePointReference endLB,
    TimePointReference endUB) {
    // endLB is the max of the abs end LB and the rel end LB + curTime.
    // endUB is the min of the abs end UB and the rel end UB + curTime.
    TimeCondition absEnd = current_node.getAbsoluteEndBounds();
    TimePoint absUB;
    if (absEnd != null)
      absUB = absEnd.getEndTime();
    else
      absUB = null;
    TimeCondition relEnd = current_node.getRelativeEndBounds();
    if (relEnd != null) {
      TimePoint relUB = relEnd.getEndTime(); // really a TimeDate*.
      TimePoint offsetUB = curTime.plus(relUB);
      if (absUB == null || offsetUB.lt(absUB))
        endUB.tp = offsetUB;
      else {
        //delete offsetUB;
        offsetUB = null;
        endUB.tp = new TimePoint(absUB);
      }
    } else {
      if (absUB != null)
        endUB.tp = new TimePoint(absUB);
      else
        endUB.tp = null;
    }
    TimePoint absLB;
    if (absEnd != null)
      absLB = absEnd.getStartTime();
    else
      absLB = null;
    if (relEnd != null) {
      TimePoint relLB = relEnd.getStartTime(); // really a TimeDate*.
      TimePoint offsetLB = curTime.plus(relLB);
      if (absLB == null || offsetLB.gt(absLB))
        endLB.tp = offsetLB;
      else {
        //delete offsetLB;
        offsetLB = null;
        endLB.tp = new TimePoint(absLB);
      }
    } else {
      if (absLB != null)
        endLB.tp = new TimePoint(absLB);
      else
        endLB.tp = null;
    }
  }

  boolean setNodeStartConditionTimeout(
    Node current_node,
    TimePoint curTime,
    TimePoint startUB) {
    // Compare the current time against the minUB.
    //Verify.beginAtomic();
    if (startUB == null || startUB.isPlusInfinity()) {
      // It's either null or positive infinity, so it won't time out.
      // Just return without complaining.
      //Verify.endAtomic();
      return true;
    } else if (curTime.ge(startUB)) {
      // The UB is already in the past, so return false.
      //Verify.endAtomic();
      return false;
    } else {
      // Set up a timer to notice when the time passes.
      // Fail on timer wakeup.
      //Verify.endAtomic();
      timer.addTime(startUB, current_node.get_id(), true, this);
      // The timer makes a copy of the time, so we can delete it.
      return true;
    }
  }

  boolean setNodeActionTimeout(
    Node current_node,
    TimePoint curTime,
    TimePoint endUB) {
    // Compare the current time against the endUB.
    //Verify.beginAtomic();
    if (endUB == null || endUB.isPlusInfinity()) {
      // It's either null or positive infinity, so it won't time out.
      // Just return without complaining.
      //Verify.endAtomic();
      return true;
    }
    else if (curTime.ge(endUB)) {
      // The UB is already in the past, so return false.
      //Verify.endAtomic();
      return false;
    } else {
      // Set up a timer to notice when the time passes.
      // Fail on timer wakeup.
      //Verify.endAtomic();
      timer.addTime(endUB, current_node.get_id(), true, this);
      // The timer makes a copy of the time, so we can delete it.
      return true;
    }
  }

  void waitForNodeStartWindow(
    Node current_node,
    TimePoint curTime,
    TimePoint startLB)
    throws ExecutionException {
    //Verify.beginAtomic();
    if (debug && Main.print)
      System.out.println(Thread.currentThread()+
        "Executive: waitForNodeStartWindow cnode,cTime,startLB + "
          + current_node
          + ","
          + curTime
          + ","
          + startLB);
    if (startLB == null) {
      // Just return -- we're not waiting.
      //Verify.endAtomic();
      return;
    } else if (curTime.ge(startLB)) {
      if (debug && Main.print) {
        System.out.println(Thread.currentThread()+
          "Executive: waitForNodeStartWindow: already in start window startLB < curTime"
            + startLB
            + " < "
            + curTime);
      }
      // We're already in the start window, so return.
      //Verify.endAtomic();
      return;
    } else if (startLB.isPlusInfinity()) {
      if (debug && Main.print) {
        System.out.println(Thread.currentThread()+"Executive: waitForNodeStartWindow: wait Infinity");
      }
      // Wait forever.
      //Verify.endAtomic();
      waitForTimerEvent(null);
    } else {
      if (debug && Main.print) {
        System.out.println(Thread.currentThread()+
          "Executive: waitForNodeStartWindow: wait for window startLB >= curTime"
            + startLB
            + " >= "
            + curTime);
      }
      // We're waiting.
      //Verify.endAtomic();
      ExecWakeupTime wakeup =
        timer.addTime(startLB, current_node.get_id(), false, this);
      if (debug && Main.print)
        System.out.println(Thread.currentThread()+"Added timer at " + startLB + ", wakeup=" + wakeup);
      waitForTimerEvent(wakeup);
    }
  }

  void waitForTimerEvent(ExecWakeupTime wakeup) throws ExecutionException {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: waitForTimerEvent - waiting for condVar");
    while (((wakeup == null) || timerWaiter.getSavedWakeup() != wakeup)
      && checkForExecutionSignal().getException() == ExecutionException.NoException) {
      //Verify.endAtomic();
      notifyCondVarWait();
      //Verify.beginAtomic();
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: waitForTimerEvent - got possible condVar");
    }
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: got condVar");
    if ((wakeup != null) && timerWaiter.getSavedWakeup() == wakeup) {
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: clearing timer's savedWakeup");
      timerWaiter.clearSavedWakeup(false);
    }
    // TODO:  clean up the failure condition.
    ExecutionException except = checkForExecutionSignal();
    if (except.getException() != ExecutionException.NoException) {
      pendingExecSignal = ExecutionSignal.ExecSignal_Okay;
      //Verify.endAtomic();
      throw except;
    }

    // Reset pending signal.
    pendingExecSignal = ExecutionSignal.ExecSignal_Okay;
    //Verify.endAtomic();
  }

  void executeTaskAction(Task current_task, ListOfVariableBindings bindings)
    throws ExecutionException {
    //Verify.beginAtomic();
    if (debug && Main.print)
      System.out.println(Thread.currentThread()+
        "Executive: executeTaskAction(" + current_task.get_id() + ")");
    // Wait for action status to change to one of the possible termination
    // states.  Throw a failure if some other relevant event happens.

    int commandID = getNewCommandID();
    int interruptCommandID = getNewCommandID();
    if ((debug || verbose) && Main.print)
      System.out.println(Thread.currentThread()+
        "Executive: assigning commandID "
          + commandID
          + " to task "
          + current_task.get_id()
          + '\n');
    current_task.set_command_id(commandID);
    Symbol action = current_task.get_action();
    // Don't try to execute a null action.
    if (action == null) {
      //Verify.endAtomic();
      return;
    }
    ListOfDBParameters actionParameters = current_task.get_parameters();
    Symbol interruptAction = current_task.get_interrupt();
    ListOfDBParameters interruptParameters =
      current_task.get_interrupt_parameters();

    // Call the function that implements the action; return on call failure.
    //Verify.endAtomic();
    if (!callAction(action,
		    actionParameters,
		    interruptAction,
		    interruptParameters,
		    bindings,
		    commandID,
		    interruptCommandID,
		    current_task.duration,
		    current_task.succeed)) {
      if (debug || verbose)
        System.out.println(Thread.currentThread()+"Executive: got false return from callAction");
      return;
    }
    optionalWait(current_task);
    waitForCommandTermination(commandID, current_task);
  }

  void optionalWait(Task current_task) {
  }

  int /*ActionStatus*/ waitForActionTermination(
    ListOfConditions actionTerminationCondition,
    Symbol id,
    Task current_task)
    throws ExecutionException {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: waitForActionTermination(");
    DBVariable var = new DBVariable(execStatusVarString);
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: adding condition check for condition ");
    ListOfVariableBindings emptyBindings = new ListOfVariableBindings();
    //Verify.endAtomic(); 
    ExecConditionCheckerStruct checkerStruct =
      conditionChecker.addConditionCheck(
        actionTerminationCondition,
        emptyBindings,
        true,
        false,
        id,
        true);
    for (;;) {
      //Verify.beginAtomic();
      ExecConditionCheckerStruct savedWakeup;
      for (;;) {
        savedWakeup = conditionChecker.getSavedWakeup();
        if (savedWakeup == checkerStruct
	    || checkForExecutionSignal().getException() != ExecutionException.NoException) {
          break; // Out of inner for loop.
	}
	//Verify.endAtomic();
        notifyCondVarWait();
        //Verify.beginAtomic();
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: got possible condvar for action termination");
      }

      DBValue dbval = null;
      if (savedWakeup == checkerStruct) {
        ListOfVariableBindings bindings = checkerStruct.getBindingList();
        if (bindings != emptyBindings) {
          System.out.println(Thread.currentThread()+
            "Executive: Bindings before and after action termination check don't match");
        }
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: checkerStruct bindings= ");
        dbval = VariableBinding.findBindingValue(var, emptyBindings);
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: clearing condition checker's savedWakeup\n");
        // TODO: possible multithread error here -- deleting current wakeup?
        emptyBindings.deleteListElements();
        conditionChecker.deleteSavedWakeup();
      }
      ExecutionException except = checkForExecutionSignal();
      if (except.getException() != ExecutionException.NoException) {
        pendingExecSignal = ExecutionSignal.ExecSignal_Okay;
        if (debug)
          System.out.println(Thread.currentThread()+
            " Executive: Throwing Exception " + except + " from waitforActionTermination\n");
	//Verify.endAtomic();
        actionThread.abortAction(true);
        throw except;
      }
      //Verify.endAtomic();
      // Reset pending signal.
      pendingExecSignal = ExecutionSignal.ExecSignal_Okay;

      return actionResult(current_task,dbval);
    }
  }

  int actionResult(Task current_task,DBValue dbval) {
    if ((dbval != null) && dbval.isint()) {
      int result = dbval.intval();
      if (isSuccessReturnValue(result))
        return ActionStatus.ActionSuccess;
      if (isFailureReturnValue(result))
        return ActionStatus.ActionFailure;
      if (isTerminateReturnValue(result))
        return ActionStatus.ExecTerminateStatus;
    }
    // unreachable code since dbval will not be null or not be an integer
    // also the action result will always be 1 of the 3 choices here
    return ActionStatus.ActionPending;
  }

  boolean isSuccessReturnValue(int retval) {
    return (retval == ActionExecution.EXEC_RETURNVALUE_SUCCESS);
  }

  boolean isFailureReturnValue(int retval) {
    return (retval == ActionExecution.EXEC_RETURNVALUE_FAILURE);
  }

  boolean isTerminateReturnValue(int retval) {
    return (retval == ActionExecution.EXEC_RETURNVALUE_TERMINATE);
  }

  void makeActionTerminationCondition( //WV not implemented completely
    int /*ExecCmdIDType*/ cmdID,
    ListOfConditions actionTerminationCondition) {
    actionTerminationCondition = new ListOfConditions(); //WV (1)
  }

  void waitForCommandTermination(int commandID, Task current_task)
    throws ExecutionException {
    //Verify.beginAtomic();
    // Set up a condition to monitor for action termination.
    ListOfConditions actionTerminationCondition = new ListOfConditions();
    makeActionTerminationCondition(commandID, actionTerminationCondition);
    if (debug || verbose)
      System.out.println(Thread.currentThread()+"Executive: about to wait for action Termination");
    boolean failure = false;
    boolean terminate = false;
    //Verify.endAtomic();
    if (waitForActionTermination(actionTerminationCondition,
      current_task.get_id(),
      current_task)
      == ActionStatus.ExecTerminateStatus) {
      //Verify.beginAtomic();
      terminate = true;
    } else if (waitForActionTermination(
		     actionTerminationCondition,
		     current_task.get_id(),
		     current_task)
	       == ActionStatus.ActionFailure) {
      //Verify.beginAtomic();
      if (debug || verbose)
        System.out.println(Thread.currentThread()+"Executive: got ActionFailure");
      failure = true;
    }
    if (actionTerminationCondition != null) {
      // TODO: check that this doesn't clear too many conditions.
      clearNodeConditions(current_task);
      actionTerminationCondition.deleteListElements();
      //delete actionTerminationCondition;
      actionTerminationCondition = null;
    }
    //Verify.endAtomic();
    if (terminate)
      throw new ExecutionException(ExecutionException.ExecTerminate);
    else if (failure)
      throw new ExecutionException(ExecutionException.NodeFailure);
  }

  boolean callAction(
    Symbol actionName, 
    ListOfDBParameters params,
    Symbol interruptActionName,
    ListOfDBParameters interruptParams,
    ListOfVariableBindings bindings,
    int /*ExecCmdIDType*/ cmdID,
    int /*ExecCmdIDType*/ interruptCmdID,
     //AG added for testing purposes
    long duration,
    boolean succeed) {
    actionThread.doAction(true); // acquire lock   
    if (debug || verbose)
      System.out.println(Thread.currentThread()+"Executive: doAction returned");
    return true;
  }

  void executeBranch(Branch current_branch, ListOfVariableBindings bindings)
    throws ExecutionException {
    if (current_branch == null)
      return;
    ListOfOptions branchOptions = current_branch.getOptions();
    if ((branchOptions == null) || branchOptions.length() == 0)
      return;

    int bestOption = -1;
    double bestUtility = 0;
    ListOfVariableBindings bestBindings = null;
    for (int iopt = 0; iopt < branchOptions.length(); iopt++) {
      // Don't even bother checking the conditions if the utility is lower
      // than the best so far.
      if (debug)
        System.out.println(Thread.currentThread()+
          "Executive: Looking at option " + branchOptions.nth(iopt).getId());
      double optUtil = branchOptions.nth(iopt).getUtility();
      if (bestOption >= 0 && bestUtility >= optUtil)
        continue;
      // Otherwise check the conditions.
      ListOfConditions optConds = branchOptions.nth(iopt).getEligibleConditions();
      // TODO: make sure this copies everything that we'll delete.
      ListOfVariableBindings optBindings = new ListOfVariableBindings();
      bindings.copyListElements(optBindings);
      boolean conditionResult = checkConditionList(optConds, optBindings, true);
      if (conditionResult) {
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: the best option, replacing old option");
        // Then we know the utility is better because we already checked that.
        bestOption = iopt;
        bestUtility = optUtil;
        if (bestBindings != null) {
          bestBindings.deleteListElements();
          //WV delete bestBindings;
          bestBindings = null;
        }
        bestBindings = optBindings;
      } else {
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: not the best option, cleaning up");
        // Not the best.  Clean up.
        optBindings.deleteListElements();
        //WV delete optBindings;
        optBindings = null;
      }
    }
    // If we don't have a legal option, throw a failure.
    if (bestOption < 0) {
      if (debug || verbose)
        System.out.println(Thread.currentThread()+"Executive: No legal options for branch");
      throw new ExecutionException(ExecutionException.NodeFailure);
    }
    // Otherwise get the node of the option and execute it.
    Node subnode = branchOptions.nth(bestOption).getNode_nonconst();
    // Record which option we took.
    ExecutionContextObject contextObj = context.top_object();
    contextObj.setPositionInNode(bestOption);
    if (bestBindings == null) {
      System.out.println(Thread.currentThread()+"Executive: null bestBindings in executeBranch");
      bestBindings = new ListOfVariableBindings();
    }
    context.add_node(subnode, bestBindings);
    // bindings are copied when adding to the context, so we can delete here.
    bestBindings.deleteListElements();
    //WV delete bestBindings;
    bestBindings = null;
    // Recursively execute the subnode.
    executeCurrentNode();
  }

  void optionalAssert() {
  }

  void performNodeAction(Node current_node, ListOfVariableBindings bindings)
    throws ExecutionException {
    //Verify.beginAtomic();
    if (current_node == null) {
      //Verify.endAtomic();
      return;
    }
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: performNodeAction ");
    switch (current_node.getSubtype()) {
      case Node.NodeSubtypeTask :
        //Verify.endAtomic();
        optionalAssert();
        executeTaskAction((Task) current_node, bindings);
        break;
      case Node.NodeSubtypeBranch :
        //Verify.endAtomic();
        executeBranch((Branch) current_node, bindings);
        break;
      case Node.NodeSubtypeBlock :
        {
          // Push each subnode on the execution stack and then execute it.
          ListOfNodes nodelist = ((Block) current_node).getNodeList();
          ExecutionContextObject contextObj = context.top_object();
          if (nodelist != null) {
	    for (int i = 0; i < nodelist.length(); i++) {
	      if (debug)
		System.out.println(Thread.currentThread()+
				   "Executive: executing subnode " + i + " of block " + current_node.get_id());
	      if (i > 0) {
		Node top_node = context.pop_node();
		if (top_node != nodelist.nth(i - 1))
		  System.out.println(Thread.currentThread()+
				     "Top of execution stack isn't the " + (i - 1) + "th node of the block");
	      }
	      // When added to the execution context, a deep copy is done of
	      // the bindings, so there won't be conflicts afterwards.
	      context.add_node((Node) nodelist.nth(i), bindings);
	      contextObj.setPositionInNode(i);
	      // Recursively execute the subnode.
	      //Verify.endAtomic();
	      executeCurrentNode();
	    }
          }
        }
        break;
      default :
        System.out.println(Thread.currentThread()+"performNodeAction: node type undefined");
        //Verify.endAtomic();
    }
  }

  void waitForNodeWaitForConditions(Node current_node, 
				    ListOfVariableBindings bindings) 
    throws ExecutionException {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: waitForNodeWaitForConditions");
    if (current_node == null) {
      //Verify.endAtomic();
      return;
    }
    ListOfConditions conds = current_node.get_wait_for_conditions();
    if (conds == null) {
      //Verify.endAtomic();
      return;
    }
    //Verify.endAtomic();
    ExecConditionCheckerStruct checkerStruct =
      conditionChecker.addConditionCheck(
        conds,
        bindings,
        true,
        false,
        current_node.get_id(),
        true);
    //Verify.beginAtomic();
    // Now wait, but watch for other events.
    for (;;) {
      ExecConditionCheckerStruct savedWakeup;
      for (;;) {
        savedWakeup = conditionChecker.getSavedWakeup();
        if (savedWakeup == checkerStruct
          || checkForExecutionSignal().getException() != ExecutionException.NoException)
          break; // Out of inner for loop.
        if (debug)
          System.out.println(Thread.currentThread()+
            "Executive: waitForNodeWaitForConditions waiting for condVar");
        //Verify.endAtomic();
        notifyCondVarWait();
        //Verify.beginAtomic();
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: got possible condvar for action termination\n");
      }
      if (savedWakeup == checkerStruct) {
        ListOfVariableBindings checkerBindings = checkerStruct.getBindingList();
        if (bindings != checkerBindings) {
          System.out.println(Thread.currentThread()+
            "Executive: Bindings before and after wait-for-condition check don't match\n");
        }
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: checkerStruct bindings= ");
        if (debug)
          System.out.println(Thread.currentThread()+"Executive: clearing condition checker's savedWakeup\n");
        conditionChecker.deleteSavedWakeup();
        // Return because we matched the needed condition.
        //Verify.endAtomic();
        return;
      }

      ExecutionException except = checkForExecutionSignal();
      if (except.getException() != ExecutionException.NoException) {
        pendingExecSignal = ExecutionSignal.ExecSignal_Okay;
        if (debug || verbose)
          System.out.println(Thread.currentThread()+
            "Executive: Throwing Exception "
              + except
              + " from waitforNodeWaitForConditions\n");
        //Verify.endAtomic();
        throw except;
      }
      // Reset pending signal.
      pendingExecSignal = ExecutionSignal.ExecSignal_Okay;
    }
  }

  /* private */
  boolean checkConditionListBody(
    ListOfConditions conditionList,
    ListOfVariableBindings bindings) {


    boolean result = false;
    DBVariable timeRefVar = new DBVariable(execReferenceTimeVarString);
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: looking for binding of ");
    DBValue timeRefVal = VariableBinding.findBindingValue(timeRefVar, bindings);
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: got binding value ");
    double timeRefDouble;
    boolean timeRefFound = false;
    if ((timeRefVal != null) && timeRefVal.isnumeric()) {
      timeRefDouble = timeRefVal.doubleval();
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: setting reference time to ");
      timeRefFound = true;
    } else {
      timeRefFound = false;
      timeRefDouble = 0.0;
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: setting reference time to 0\n");
    }
    TimePoint referenceTime = new TimePoint(timeRefDouble);
    TimePoint curTime = timer.getCurrentTime();

    boolean found_result = false;
    result = true; // True if we get through the list w/o a false.
    for (int i = 0; i < conditionList.length(); i++) {
      if (debug)
        System.out.println(Thread.currentThread()+"checkConditionList: i=");
      switch (conditionList.nth(i).get_conditionsubtype()) {
        case ConditionSubtype.DatabaseConditionSubtype :
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: DatabaseCondition\n");
          if (!db.DBQuery(
            conditionList.nth(i), bindings, false)) {
            result = false;
            found_result = true;
            if (debug)
              System.out.println(Thread.currentThread()+"checkConditionList: failing on condition ");
          }
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: after DBQuery, result=" + result);
          break;
        case ConditionSubtype.ResourceConditionSubtype :
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: ResourceCondition\n");
          if (!db.ResourceQuery(
            conditionList.nth(i), bindings, false)) {
            result = false;
            found_result = true;
            if (debug)
              System.out.println(Thread.currentThread()+"checkConditionList: failing on condition ");
          }
          if (debug)
            System.out.println(Thread.currentThread()+
              "checkConditionList: after ResourceQuery, result="
                + result
                + ", found_result="
                + found_result);
          break;
        case ConditionSubtype.FunctionConditionSubtype :
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: FunctionCondition\n");
          if (!(db.FunctionQuery(
            conditionList.nth(i), bindings, functionTable))) {
            result = false;
            found_result = true;
            if (debug)
              System.out.println(Thread.currentThread()+"checkConditionList: failing on condition ");
          }
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: after fn test, result=" + result);
          break;
        case ConditionSubtype.TimeConditionSubtype :
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: TimeCondition\n");
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: Warning: using default referenceTime");
          if (!((TimeCondition) (conditionList.nth(i)))
            .satisfied(bindings, referenceTime, curTime)) {
            if (debug)
              System.out.println(Thread.currentThread()+"Got a false return from TimeCondition::satisfied\n");
            result = false;
            found_result = true;
            if (debug)
              System.out.println(Thread.currentThread()+"checkConditionList: failing on condition ");
          }
          if (debug)
            System.out.println(Thread.currentThread()+"checkConditionList: after time check, result=" + result);
          break;
        default :
          System.out.println(Thread.currentThread()+"condition unknown type\n");
      }
      if (found_result)
        break;
    }
    return result;
  }

  public boolean checkConditionList(
    ListOfConditions conditionList,
    ListOfVariableBindings bindings,
    boolean lockDBMutex) {

    //Verify.beginAtomic();

    boolean result = false;
    if (debug)
      System.out.println(Thread.currentThread()+"checkConditionList: conditionList length = ");
    if (lockDBMutex) {
      //Verify.endAtomic();
      synchronized (db) {
        //Verify.beginAtomic();
        result = checkConditionListBody(conditionList, bindings);
      }
    } else {
      result = checkConditionListBody(conditionList, bindings);
    }
    if (debug)
      System.out.println(Thread.currentThread()+"Returning " + result + " from checkConditionList\n");
    //Verify.endAtomic();
    return result;
  }

  void setupNodeMaintainConditions(Node current_node, 
				   ListOfVariableBindings bindings) {
    //Verify.beginAtomic();
    if (verbose)
      System.out.println(Thread.currentThread()+"Executive: setupNodeMaintainConditions");
    if (current_node == null) {
      //Verify.endAtomic();
      return;
    }
    // TODO: using the same bindings here could cause weirdness if variables
    // are re-used.  Should make a copy before calling this function.
    ListOfConditions conds = current_node.get_maintain_conditions();
    if (conds == null) {
      //Verify.endAtomic();
      return;
    }
    //Verify.endAtomic();
    conditionChecker.addConditionCheck(
      conds,
      bindings,
      false,
      true,
      current_node.get_id(),
      true);
    // We won't use the struct, because the checker will signal a failure
    // directly.
  }

  void clearNodeTimeouts(Node current_node) {
    if (current_node == null)
      return;
    timer.removeTime(current_node.get_id());
  }

  void clearNodeConditions(Node current_node) {
    if (current_node == null)
      return;
    conditionChecker.removeConditionCheck(current_node.get_id(), true);
  }

  boolean checkNodePreconditions(
    Node current_node,
    ListOfVariableBindings bindings) {
    if (current_node == null)
      return false;
    if (debug)
      System.out.println(Thread.currentThread()+"Executive::checkNodePreconditions(");
    boolean result =
      checkConditionList(current_node.get_start_conditions(), bindings, true);
    return result;
  }

  boolean checkNodePostconditions(
    Node current_node,
    ListOfVariableBindings bindings) {
    //Verify.beginAtomic();
    if (current_node == null) {
      //Verify.endAtomic();
      return false;
    }
    if (debug)
      System.out.println(Thread.currentThread()+"Executive::checkNodePostconditions(");
    //Verify.endAtomic();
    boolean result =
      checkConditionList(current_node.get_end_conditions(), bindings, true);
    return result;
  }

  void condition(Node current_node, int i) {
      if (i == 0)
        System.out.println(Thread.currentThread()+current_node.get_id().toString() + " START");
      else if (i == 1) 
        System.out.println(Thread.currentThread()+current_node.get_id().toString() + " SUCCESS");
      else if (i == -1)
        System.out.println(Thread.currentThread()+current_node.get_id().toString() + " FAILURE");
      else 
	System.out.println(Thread.currentThread()+"UNKNOWN condition!!!!");
  }

  void executeCurrentNode() throws ExecutionException {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: executeCurrentNode");
    if (context == null) {
      System.out.println(Thread.currentThread()+"No execution context!");
      //Verify.endAtomic();
      return;
    }
    Node current_node = context.get_currentNode();
    ExecutionContextObject current_object = context.top_object();
    if (current_node == null) {
      System.out.println(Thread.currentThread()+"No current node!");
      //Verify.endAtomic();
      return;
    }
    if (debug || verbose)
      System.out.println(Thread.currentThread()+"Executive: executeCurrentNode");
    TimePoint startLB = null;
    TimePoint startUB = null;
    ListOfVariableBindings bindings = context.top_bindings();
    ListOfVariableBindings bindingsForMaintainCheck = null;
    try {
      TimePoint curTime =
        setAndAssertNodeTime(
          NodeTimeType.NodeTransitionTime,
          current_node.get_id(),
          current_object);
      // TODO: clean up these memory allocations.
      DBVariable var = new DBVariable(execReferenceTimeVarString);
      VariableBinding oldBinding =
        VariableBinding.findBinding_nonconst(var, bindings);
      DBValue val = new DBValue(curTime.doubleval());
      if (oldBinding != null) {
        oldBinding.updateVariableBinding(val);
      } else {
        VariableBinding binding = new VariableBinding(var, val);
        bindings.push(binding);
      }
      if (debug)
        System.out.println(Thread.currentThread()+
          "Executive: bindings after adding reference time (transition time): ");

      // Side effect: sets startLB and startUB.
      TimePointReference tprStartLB = new TimePointReference();
      TimePointReference tprStartUB = new TimePointReference();
      tprStartLB.tp = startLB;
      tprStartUB.tp = startUB;
      getStartTimeBounds(current_node, curTime, tprStartLB, tprStartUB);
      startLB = tprStartLB.tp;
      startUB = tprStartUB.tp;
      if (debug && Main.print)
        System.out.println(Thread.currentThread()+"StartLB 2 = " + startLB);
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: start time bounds [");
      //  TODO: implement the functions below.

      // Set up start condition upper bound.
      //Verify.endAtomic();
      if (!setNodeStartConditionTimeout(current_node, curTime, startUB)) {
        //Verify.beginAtomic();
        // Too late already.
        if (debug || verbose)
          System.out.println(Thread.currentThread()+"Executive: too late for start time, throwing failure");
        signaledNodeID = current_node.get_id(); // Current node only.

        //Verify.endAtomic();
        throw new ExecutionException(ExecutionException.NodeFailure);
      }
      // Wait for start condition lower bound.
      if (debug || verbose)
        System.out.println(Thread.currentThread()+"Executive: waiting for node start lower bound ");
      waitForNodeStartWindow(current_node, curTime, startLB);
      if (debug || verbose)
        System.out.println(Thread.currentThread()+"Executive: done waiting for node start lower bound");

      waitForNodeWaitForConditions(current_node, bindings);
      clearNodeTimeouts(current_node);
      if (!checkNodePreconditions(current_node, bindings))
        throw new ExecutionException(ExecutionException.ConditionFailure);
      //Verify.beginAtomic();
      bindingsForMaintainCheck = new ListOfVariableBindings();
      bindings.copyListElements(bindingsForMaintainCheck);
      //Verify.endAtomic();
      setupNodeMaintainConditions(current_node, bindings);
      //Verify.beginAtomic();
    } catch (ExecutionException except) {
      //Verify.beginAtomic();
      switch (except.getException()) {
        case ExecutionException.NodeFailure :
        case ExecutionException.PlanAbort :
        case ExecutionException.ConditionFailure :
          {
            //Verify.endAtomic();
            clearNodeTimeouts(current_node);
            clearNodeConditions(current_node);
            if (bindingsForMaintainCheck != null) {
              bindingsForMaintainCheck.deleteListElements();
            }
            // Set start time and end time to now.
            //Verify.beginAtomic();
            setAndAssertNodeTime(
              NodeTimeType.NodeStartTime,
              current_node.get_id(),
              current_object);
            setAndAssertNodeTime(
              NodeTimeType.NodeEndTime,
              current_node.get_id(),
              current_object);

            // TODO: Check for continue on failure and signaledNodeID.
            boolean contval =
              current_node.get_continue_on_failure().boolval(
                context.getInheritedContinueOnFailure());
            if (debug)
              System.out.println(Thread.currentThread()+
                "Executive: Node continue_on_failure is "
                  + contval
                  + " for Node "
                  + current_node);
            if ((except.getException() == ExecutionException.NodeFailure
              || except.getException() == ExecutionException.ConditionFailure)
              && contval) {
              if (debug || verbose)
                System.out.println(Thread.currentThread()+"Executive: stopping propagation of NodeFailure");
              // still need to stop execution of this node.
              //Verify.endAtomic();
              return;
            } else {
              if (debug || verbose)
                System.out.println(Thread.currentThread()+
                  "Executive: propagating ExecutionException (NodeFailure or PlanAbort or ConditionFailure) ");
              //Verify.endAtomic();
              throw except;
            }
          }
        default :
	  if (debug || verbose)
            System.out.println(Thread.currentThread()+"Executive: propagating (unknown) ExecutionException ");
          clearNodeTimeouts(current_node);
          clearNodeConditions(current_node);
          if (bindingsForMaintainCheck != null) {
            bindingsForMaintainCheck.deleteListElements();
          }
          throw except;
      }
    }

    TimePoint endLB = null;
    TimePoint endUB = null;
    try {
      TimePoint curTime =
        setAndAssertNodeTime(
          NodeTimeType.NodeStartTime,
          current_node.get_id(),
          current_object);

      // TODO: clean up these memory allocations.
      DBVariable var = new DBVariable(execReferenceTimeVarString);
      VariableBinding oldBinding =
        VariableBinding.findBinding_nonconst(var, bindings);
      DBValue val = new DBValue(curTime.doubleval());
      if (oldBinding != null) {
        oldBinding.updateVariableBinding(val);
      } else {
        VariableBinding binding = new VariableBinding(var, val);
        bindings.push(binding);
      }
      if (debug)
        System.out.println(Thread.currentThread()+
          "Executive: bindings after adding reference time (start time): ");
      TimePointReference tprEndLB = new TimePointReference();
      TimePointReference tprEndUB = new TimePointReference();
      tprEndLB.tp = endLB;
      tprEndUB.tp = endUB;
      getEndTimeBounds(current_node, curTime, tprEndLB, tprEndUB);
      endLB = tprEndLB.tp;
      endUB = tprEndUB.tp;
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: end time bounds [");
      //Verify.endAtomic();
      if (!setNodeActionTimeout(current_node, curTime, endUB)) {
        // Too late, fail.
        signaledNodeID = current_node.get_id(); // Just current node.
        throw new ExecutionException(ExecutionException.NodeFailure);
      }
      // The node action is either an action or a subnode.
      condition(current_node,0);
      performNodeAction(current_node, bindings);
      if (debug)
        System.out.println(Thread.currentThread()+"Executive: clearing Node conditions after execution\n");
      clearNodeTimeouts(current_node);
      clearNodeConditions(current_node);
      if (bindingsForMaintainCheck != null) {
        bindingsForMaintainCheck.deleteListElements();
      }
      if (debug)
        System.out.println(Thread.currentThread()+
          "Executive: done clearing Node conditions after execution\n");
      if (!checkNodePostconditions(current_node, bindings))
        throw new ExecutionException(ExecutionException.ConditionFailure);
      condition(current_node,1);
    } catch (ExecutionException except) {
      switch (except.getException()) {
        case ExecutionException.PlanAbort :
	  condition(current_node,-1);
          Main.testSuite.check(current_node.get_id().toString(), -1);
          if (debug || verbose)
            System.out.println(Thread.currentThread()+"Executive: Caught NodeFailure during execution of");
          clearNodeTimeouts(current_node);
          clearNodeConditions(current_node);
          //Verify.beginAtomic();
          setAndAssertNodeTime(
            NodeTimeType.NodeEndTime,
            current_node.get_id(),
            current_object);
          //Verify.endAtomic();
          if (debug || verbose)
            System.out.println(Thread.currentThread()+"Executive: propagating PlanAbort\n");
          throw except;
        case ExecutionException.NodeFailure :
        case ExecutionException.ConditionFailure :
          if (debug || verbose)
            System.out.println(Thread.currentThread()+
              "Executive: Caught NodeFailure or ConditionFailure during execution of "
                + current_node);
          clearNodeTimeouts(current_node);
          clearNodeConditions(current_node);
          //Verify.beginAtomic();
          setAndAssertNodeTime(
            NodeTimeType.NodeEndTime,
            current_node.get_id(),
            current_object);
          //Verify.endAtomic();
          // Pass on a 
          if (!current_node.get_continue_on_failure().boolval(context.getInheritedContinueOnFailure())) {
	     condition(current_node,-1);
            if(debug || verbose)
              System.out.println(Thread.currentThread()+"Executive: propagating NodeFailure");
            // TODO: check continue on failure and signaledNodeID.
            throw except;
          } else {
	      if (debug || verbose)
              System.out.println(Thread.currentThread()+
                "Executive: NOT propagating NodeFailure since continue on failure is not inherited and set to true");
          }
          condition(current_node,1);
        default :
	  condition(current_node,-1);
          clearNodeTimeouts(current_node);
          clearNodeConditions(current_node);
          //Verify.beginAtomic();
          setAndAssertNodeTime(
            NodeTimeType.NodeEndTime,
            current_node.get_id(),
            current_object);
          //Verify.endAtomic();
          if (debug || verbose)
            System.out.println(Thread.currentThread()+
              "Executive: propagating (unknown) ExecutionException at end of node execution");
          throw except;
      }
    }
    //Verify.beginAtomic();
    setAndAssertNodeTime(
      NodeTimeType.NodeEndTime,
      current_node.get_id(),
      current_object);
    //Verify.endAtomic();
  }

  boolean emptyExecutionContext() {
    return ((context == null) || context.emptyExecutionContext());
  }

  public int contextDepth(Symbol id) {
    // Return the depth of id in context stack.
    if (context == null)
      return -1;
    return (context.contextDepth(id));
  }

  /* private */
  boolean body(int sig) {
    boolean sigChanged = false;
    if (sig != pendingExecSignal) {
      switch (sig) {
        case ExecutionSignal.ExecSignal_PlanAbort :
          pendingExecSignal = sig;
          sigChanged = true;
          break;
        case ExecutionSignal.ExecSignal_Timeout :
        case ExecutionSignal.ExecSignal_ConditionFailure :
          if (pendingExecSignal != ExecutionSignal.ExecSignal_PlanAbort) {
            pendingExecSignal = sig;
            sigChanged = true;
          }
          break;
        default :
          System.out.println(Thread.currentThread()+"Reporting ExecutionSignal ");
      }
    }
    return sigChanged;
  }

  public void reportExecutionSignal(int /*ExecutionSignal*/ sig, boolean lock) {
    boolean sigChanged;
    if (debug || verbose)
      System.out.println(Thread.currentThread()+"Reporting execution signal");
    if (lock) {
      synchronized (this) {
        sigChanged = body(sig);
        if (sigChanged) {
          notifyCondVarSignal();
        }
      }
    } else {
      sigChanged = body(sig);
      if (sigChanged) {
        notifyCondVarSignal();
      }
    }
  }

  public void notifyCondVarSignal() {
    notifyAll();
  }

  void notifyCondVarWait() {
    try {
      wait();
    } catch (InterruptedException e) {
    };
  }

  Database getDB() {
    return db;
  }

  void displayExecutionStatus() {
  }
  void resetExecutionSignal() {
    pendingExecSignal = ExecutionSignal.ExecSignal_Okay;
  }

  void setPendingPlanFile(String planfile) {
    pendingPlanFile = planfile;
  }

  String getPendingPlanFile() {
    return pendingPlanFile;
  }

  CommandTable getCommandTable() {
    return commandTable;
  }
  FunctionTable getfunctionTable() {
    return functionTable;
  }

  public void start_Executive_and_Timer() {
    start(); //WV runThreads();
    timer.start();
  }
  public ExecConditionChecker getConditionChecker() {
    return conditionChecker;
  }
  public ExecTimer getTimer() {
    return timer;
  }
  public ExecTimerWaiter getTimerWaiter() {
    return timerWaiter;
  }

  // topLevelLoop
  public void run() {
    synchronized (this) {
      //Verify.beginAtomic();
      boolean skip_once = true;
      boolean terminate = false;
      while (!terminate) {
        try {
          // Look for new plan.
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: run - check for new plan");
          while (pendingPlanFile == null) { // Wait for a change.
            if (debug)
              System.out.println(Thread.currentThread()+"Executive: run - no plan wait");
            //Verify.endAtomic();
            notifyCondVarWait();
            //Verify.beginAtomic();
          }
          if (debug)
            System.out.println(Thread.currentThread()+"Executive: run - plan signaled");
          resetExecutionSignal();
          pendingPlanFile = null;
          if (Main.testSuite != null) {
            Main.testSuite.reset();
            while (Main.testSuite.hasNext()) {
              plan = Main.testSuite.next();
              //Verify.endAtomic();
              changePlan(plan, PlanInstallType.PlanInstall_Replace);
              //Verify.beginAtomic();
            }
            terminate = true;
          } else
            System.out.println(Thread.currentThread()+"Executive: No test suite");
        } catch (ExecutionException except) {
          switch (except.getException()) {
            case ExecutionException.ExecTerminate :
              terminate = true;
              break;
            default :
              System.out.println(Thread.currentThread()+"Unknown ExecutionException at top level" + except);
          }
        }
      }
      System.out.println(Thread.currentThread()+"Executive is DONE");
      System.exit(0);
      // We should never get here unless we're done.
      // TODO: unlock the mutex in the thread cleanup code.
    }
  }

  Node lookForNewPlan() {
    return Planner.plan1();
  }

  void changePlan(Node newplan, int /*PlanInstallType*/ install)
    throws ExecutionException {
    //Verify.beginAtomic();
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: changePlan");
    switch (install) {
      case PlanInstallType.PlanInstall_Merge :
        System.out.println(Thread.currentThread()+"Plan merging not implemented -- replacing");
      case PlanInstallType.PlanInstall_Append :
        if (!emptyExecutionContext()) {
          System.out.println(Thread.currentThread()+
            "Trying to change plans with Append mode when execution context is nonempty; Replacing plan\n");
        }
        // Drop through to replace, since they work the same way.
      case PlanInstallType.PlanInstall_Replace :
        if (!emptyExecutionContext()) {
          context.clearExecutionContext();
        }
        execStatus = ExecStatus.ExecStatus_Running;
        initializePlan(newplan);
        initializeContext();
        //Verify.endAtomic();
        executePlan(false);
        //Verify.beginAtomic();
        break;
      default :
        System.out.println(Thread.currentThread()+"Unknown installation option:");
    }
    //Verify.endAtomic();
  }

  ExecutionException checkForExecutionSignal() {
    if (debug)
      System.out.println(Thread.currentThread()+"Executive: checkForExecutionSignal " + pendingExecSignal);
    switch (pendingExecSignal) {
      case ExecutionSignal.ExecSignal_PlanAbort :
        return new ExecutionException(ExecutionException.PlanAbort);
      case ExecutionSignal.ExecSignal_Timeout :
      case ExecutionSignal.ExecSignal_ConditionFailure :
        return new ExecutionException(ExecutionException.NodeFailure);
      case ExecutionSignal.ExecSignal_Okay :
        return new ExecutionException(ExecutionException.NoException);
      default :
        System.out.println(Thread.currentThread()+"Executive: Unknown pendingExecSignal " + pendingExecSignal);
        return new ExecutionException(ExecutionException.NoException);
    }
  }
}
