package rover;

//import gov.nasa.jpf.jvm.Verify;
import java.util.ArrayList;
import java.util.Iterator;

class Node {

  final static boolean debug = Main.print;

  public final static int NodeSubtypeUndefined 	= 0;
  public final static int NodeSubtypeTask 	= 1;
  public final static int NodeSubtypeBranch    	= 2;
  public final static int NodeSubtypeBlock  	= 3;

  static Symbol id_default = Symbol.NullSymbol;
  static String comment_default = new String("");
  static ListOfConditions start_conditions_default = null;
  static ListOfConditions wait_for_conditions_default = null;
  static ListOfConditions maintain_conditions_default = null;
  static ListOfConditions end_conditions_default = null;
  static ListOfNodeIDs order_constraints_default = null; 
  static double utility_default = 0;
  static double tail_utility_default = 0;
  static double probability_default = 0;

  static ContinueType continue_on_failure_default = 
    new ContinueType(ContinueType.Inherited);
  static ContinueType default_continue_on_failure_default =
    new ContinueType(ContinueType.Inherited);
  static ListOfAlternatePlans alternate_plans_default = null;


  Symbol id;
  String comment;
  ListOfConditions start_conditions;
  ListOfConditions wait_for_conditions;
  ListOfConditions maintain_conditions;
  ListOfConditions end_conditions;
  //WV ListOfNodeIDs* order_constraints;
  double utility;
  double tail_utility;
  double probability;
  ListOfAlternatePlans alternate_plans;

  ContinueType continue_on_failure;
  ContinueType default_continue_on_failure;

  TimeCondition absoluteStartBounds;
  TimeCondition relativeStartBounds;
  TimeCondition absoluteEndBounds;
  TimeCondition relativeEndBounds;

  int /* NodeSubtype */ subtype;

  public String toString() {
    return "" + id;
  }

  public Node() {
    subtype = NodeSubtypeUndefined;
    id = Symbol.NullSymbol;
    continue_on_failure = new ContinueType(continue_on_failure_default);
    default_continue_on_failure = 
      new ContinueType(default_continue_on_failure_default);
    utility = 0;//utility_default;
    tail_utility = 0;//tail_utility_default;
    probability = 0;//probability_default;
    //WV     initCompiledConditions();
    //WV initExtra();
  }

  public Node(int /*NodeSubtype*/ xsubtype,
	      Symbol xid,
	      String xcomment,
	      ListOfConditions xstart_conditions,
	      ListOfConditions xwait_for_conditions,
	      ListOfConditions xmaintain_conditions,
	      ListOfConditions xend_conditions,
	      ListOfNodeIDs xorder_constraints,
	      double xutility, double xtail_utility, double xprobability,
	      ContinueType xcontinue_on_failure, 
	      ContinueType xdefault_continue_on_failure,
	      ListOfAlternatePlans xalternate_plans) {
    subtype = xsubtype;
    id = xid;
    //comment = xcomment;
    start_conditions = xstart_conditions;
    wait_for_conditions = xwait_for_conditions;
    maintain_conditions = xmaintain_conditions;
    end_conditions = xend_conditions;
    //WV order_constraints = xorder_constraints;
    utility = xutility; 
    tail_utility = xtail_utility; 
    probability = xprobability;
    continue_on_failure = xcontinue_on_failure; 
    default_continue_on_failure = xdefault_continue_on_failure;
    alternate_plans = xalternate_plans;
    initCompiledConditions();
    //initExtra();
  }

  void initCompiledConditions() {
    absoluteStartBounds = absoluteEndBounds = relativeStartBounds =
      relativeEndBounds = null;
    //WV    waitForResourceConditions = startResourceConditions =
    //WV  maintainResourceConditions = endResourceConditions = 0;
    //WV unpredictableWaitForConditions = unpredictableStartConditions =
    //WV  unpredictableMaintainConditions = unpredictableEndConditions = 0;
  }

  public int /* NodeSubtype */ getSubtype() {
    return subtype;
    /* WV
    int result = //Verify.random(3); // don't want 0 since it is undefined.
    if (result == 0) return 1;
    return result;
    */
  }
 
  public ContinueType get_continue_on_failure() {
    return continue_on_failure; 
  }
  
  public ContinueType get_default_continue_on_failure() {
    return default_continue_on_failure; 
  }

  public Symbol get_id() {
    return id; //new Symbol();
  }

  public ListOfConditions get_start_conditions() {
    return start_conditions;
  }

  public ListOfConditions get_end_conditions() {
    return end_conditions;
  }

  public ListOfConditions get_maintain_conditions() {
    return maintain_conditions;
  }

  public ListOfConditions get_wait_for_conditions() {
    return wait_for_conditions;
  }

  TimeCondition getAbsoluteStartBounds() { 
    return absoluteStartBounds; 
  }
  TimeCondition getRelativeStartBounds() {
    return relativeStartBounds; 
  }
  TimeCondition getAbsoluteEndBounds() { 
    return absoluteEndBounds; 
  }
  TimeCondition getRelativeEndBounds() {
    return relativeEndBounds; 
  }


  void compileStartConditions() {
    absoluteStartBounds = new TimeCondition();
    relativeStartBounds = new TimeCondition();
    //start conditions must be set by now
    ListOfConditions conds = get_start_conditions();
    compileConditions(conds, absoluteStartBounds, relativeStartBounds);
    //WV startResourceConditions = compileResourceConditions(conds);
    //WV  unpredictableStartConditions = countUnpredictableConditions(conds);
    //WV ListOfConditions waitForConds = get_wait_for_conditions();
    //WV     waitForResourceConditions = compileResourceConditions(waitForConds);
    //WV   unpredictableWaitForConditions = 
    //WV    countUnpredictableConditions(waitForConds);
}

  void compileEndConditions() {
    absoluteEndBounds = new TimeCondition();
    relativeEndBounds = new TimeCondition();
    // end conditions must be set by now
    ListOfConditions conds = get_end_conditions();
    compileConditions(conds, absoluteEndBounds, relativeEndBounds);
    //WV  endResourceConditions = compileResourceConditions(conds);
    //WV  unpredictableEndConditions = countUnpredictableConditions(conds);
  }

  void compileConditions(ListOfConditions conds, 
			TimeCondition absoluteBounds,
			TimeCondition relativeBounds) {

    if (debug && Main.print) 
      System.out.println("compileConditions " + conds);

    if (conds == null)
      return;

    for (Iterator i =  conds.getIterator(); i.hasNext();) {
      TimeCondition cond = (TimeCondition)i.next();
      if (cond.get_conditionsubtype() == 
	  Condition.TimeConditionSubtype) {
	if (debug && Main.print)
	  System.out.println("Looking at time condition " + cond);
	TimeDate cond_stime = cond.getStartTime();
	TimeDate cond_etime = cond.getEndTime();
	// Now compare each of these against current bounds.
	if (cond_stime != null) {
	  if (cond_stime.isrelative()) {
	    if (debug)
	      System.out.println("Relative condition lower bound");
	    // check against relative lower bound.
	    TimeDate rel_lbound = relativeBounds.getStartTime();
	    if ((rel_lbound == null) || cond_stime.gt(rel_lbound))
	      relativeBounds.setStartTime(new TimeDate(cond_stime));
	  } else {
	    if (debug)
	      System.out.println("Absolute condition lower bound");
	    // check against absolute lower bound.
	    TimeDate abs_lbound = absoluteBounds.getStartTime();
	    if ((abs_lbound == null) || cond_stime.gt(abs_lbound))
	      absoluteBounds.setStartTime(new TimeDate(cond_stime));
	  }
	}
	if (cond_etime != null) {
	  if (cond_etime.isrelative()) {
	    if (debug)
	      System.out.println("Relative condition upper bound ");
	    // check against relative upper bound.
	    TimeDate rel_ubound = relativeBounds.getEndTime();
	    if ((rel_ubound == null) || cond_etime.lt(rel_ubound))
	      relativeBounds.setEndTime(new TimeDate(cond_etime));
	  } else {
	    if (debug)
	      System.out.println("Absolute condition upper bound ");
	    // check against absolute upper bound.
	    TimeDate abs_ubound = absoluteBounds.getEndTime();
	    if ((abs_ubound == null) || cond_etime.lt(abs_ubound))
	      absoluteBounds.setEndTime(new TimeDate(cond_etime));
	  }
	}
      }
    }
  }
}
