package rover;
class TimeInterval {
  private TimePoint startTime;
  private TimePoint endTime;
  
  public TimeInterval() {
    // TimePoint changed to TimeDate to fix casting problem
    startTime = new TimeDate(TimePoint.MinusInfinity);
    endTime = new TimeDate(TimePoint.PlusInfinity);
  }

  public TimeInterval(TimePoint stime, TimePoint etime) {
    // TimePoint changed to TimeDate to fix casting problem
    if (stime != null)
      startTime = stime;
    else
      startTime = new TimeDate(TimePoint.MinusInfinity);
    if (etime != null)
      endTime = etime;
    else
      endTime = new TimeDate(TimePoint.PlusInfinity);
  }
  
  public TimeInterval(TimeInterval ti) {
    startTime = new TimePoint(ti.startTime);
    endTime = new TimePoint(ti.endTime);
  }

  public TimePoint getStartTime() { 
    return startTime; 
  }
  public TimePoint getEndTime() { 
    return endTime; 
  }
  
  void setStartTime(TimePoint stime) { 
    //WV if (startTime != null)
    //WV   delete startTime;
    startTime = stime;
  }
  void setEndTime(TimePoint etime) {
    //WV     if (endTime)
    //WV       delete endTime;
    endTime = etime;
  }
  
  void shift(TimePoint timeToShift) {
    startTime.increment(timeToShift);
    endTime.increment(timeToShift);
  }

  boolean containsTime(TimePoint tp) {
    return (((startTime == null) || startTime.le(tp)) && 
	    ((endTime == null)|| endTime.ge(tp)));
  }

  boolean containsTimeInterval(TimeInterval ti) {
    return (containsTime(ti.getStartTime()) &&
	    containsTime(ti.getEndTime())); }

  boolean overlapsTimeInterval(TimeInterval ti) {
    return (containsTime(ti.getStartTime()) ||
	    containsTime(ti.getEndTime()) ||
	    ti.containsTime(getStartTime()) ||
	    ti.containsTime(getEndTime()));
  }

  boolean equal(TimeInterval ti) {
    return ((((startTime == null) && (ti.getStartTime() == null)) ||
	     ((startTime != null) && (ti.getStartTime() != null) && 
	      startTime.eq(ti.getStartTime()))) &&
	    (((endTime == null) && (ti.getEndTime() == null)) ||
	     ((endTime != null) && (ti.getEndTime() != null) && 
	      endTime.eq(ti.getEndTime()))));
  }

  TimePoint intervalSize() {
    return (getEndTime().minus(getStartTime()));
}

}
