package rover;

class TimePoint implements Comparable {

  Exception e;

  static final int SymbolicTimeUndefined = 0;
  static final int MinusInfinity         = 1;
  static final int PlusInfinity          = 2;

  static final int tpUNKNOWN = 0;
  static final int tpLT      = 1;
  static final int tpEQ      = 2;
  static final int tpGT      = 3;

  //used when c++ requires multiple vales returned via var-params
  class returnType {
    long s;
    int ms;
    public returnType(long xs, int xms) {
      s = xs;
      ms = xms;
    }
    long getSecs() {
      return s;
    }
    int getmSecs() {
      return ms;
    }
  }

  public String toString() {
    return "TimePoint(" + seconds + "," + milliseconds + ")";
  }

  public TimePoint(long secs, int msec) {
    numeric_init(secs, msec); 
  }

  public TimePoint(long val) {
    numeric_init(val,0); 
  }


  public TimePoint() {
    int val = 0;
    numeric_init((long) val,0); 
  }

  public TimePoint(int val, int val2) { 
    //WV had to change this to make it different from the Symbolic case
    //WV val2 will just be ignored;
    numeric_init((long) val,0); 
  }

  public TimePoint(double val) {
    numeric_init(((long) val),
		 ((int) ( ((val - ((long) val)) * 1000) + 0.5)));
  }

  public TimePoint(int /*SymbolicTimePoint*/ sym) { 
    symbolic_init(sym); 
  }

  void initTimePointFromTimePoint(TimePoint tpoint,
				  int/*SymbolicTimePoint*/ defaultsymtime) {
    if (tpoint != null) {
      if (tpoint.issymbolic()) {
	symbolic_init(tpoint.symbolicval());
      } else {
	// This has as a side effect to set the two variables.
	//WV tpoint.numericval(seconds, milliseconds);
	seconds = tpoint.numericvalSecs();
	milliseconds = tpoint.numericvalmSecs();
	issymbolictime = false;
      }
    } else {
      symbolic_init(defaultsymtime);
    }
  }

  public TimePoint(TimePoint tpoint, 
		   int /*SymbolicTimePoint*/ defaultsymtime) {
    initTimePointFromTimePoint(tpoint, defaultsymtime);
  }

  public TimePoint(TimePoint tpoint) {
    initTimePointFromTimePoint(tpoint, SymbolicTimeUndefined);
  }



  boolean issymbolictime;
  //Symbolic field
  int /*SymbolicTimePoint*/ symtime;
  // Numeric fields

  // Number of seconds since beginning of mission.
  // Since longs are >= 32 bits, this will be good for 136-year missions
  // at the minimum.
  /*unsigned*/ long seconds;
  // Number of milliseconds since beginning of mission.
  /*unsigned*/ int milliseconds; 

  public boolean symbolic_equal(TimePoint tp) {
    return (tp.issymbolic() && tp.symtime == symtime); 
  }

  public boolean numeric_equal(TimePoint tp) {
    return (tp.isnumeric() && 
	    tp.seconds == seconds && tp.milliseconds == milliseconds);
  }

  public int /*TimePointRelation*/ numeric_relation(TimePoint tp) {
    if (!(isnumeric() && tp.isnumeric()))
      return tpUNKNOWN;
    if (seconds < tp.seconds)
      return tpLT;
    else if (seconds > tp.seconds)
      return tpGT;
    else /* seconds == tp.seconds */
      if (milliseconds < tp.milliseconds)
	return tpLT;
      else if (milliseconds > tp.milliseconds)
	return tpGT;
      else /* milliseconds == tp.milliseconds */
	return tpEQ;
  }

  public void numeric_init(long secs, int msec) {
    e = new Exception();
    issymbolictime = false;
    seconds = secs; 
    milliseconds = msec;
  }

  public void symbolic_init(int /*SymbolicTimePoint*/ sym) {
    e = new Exception();
    issymbolictime = true;
    symtime = sym;
  }

  public boolean isnumeric() { 
    return (!issymbolictime); 
  }

  public boolean issymbolic() { 
    return issymbolictime; 
  }

  public long numericvalSecs() {
    return seconds;
  }

  public int numericvalmSecs() {
    return milliseconds;
  }

  public int /*SymbolicTimePoint*/ symbolicval() { 
    return symtime; 
  }

  public double doubleval() { 
    return ( seconds + milliseconds / 1000.0); 
  }

  public boolean isPlusInfinity() { 
    return (issymbolictime && symtime == PlusInfinity);
  }

  public boolean isMinusInfinity() { 
    return (issymbolictime && symtime == MinusInfinity);
  }

  public boolean equal(TimePoint tp) {
      /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert(false);

    return ((isnumeric() && tp.isnumeric() && numeric_equal(tp)) ||
	    (issymbolic() && tp.issymbolic() && symbolic_equal(tp)));
  }

  returnType numericDiff(TimePoint tp) {
    long upsecs, lowsecs;
    int upmsecs, lowmsecs;
    lowsecs = tp.numericvalSecs();
    lowmsecs = tp.numericvalmSecs();;
    upsecs = numericvalSecs();
    upmsecs = numericvalmSecs();;
    if (upmsecs < lowmsecs) {
      upmsecs += 1000;
      upsecs--;
    }
    // Howard discovered with ESC/Java that since we do upsecs--
    // it could be that upsecs < lowsecs and that will be a problem
    return new returnType(upsecs - lowsecs,upmsecs - lowmsecs);
  }

  returnType numericSum(TimePoint tp) {
    long secs;
    int msecs;
    long srcsecs, addsecs;
    int srcmsecs, addmsecs;
    addsecs = tp.numericvalSecs();
    addmsecs = tp.numericvalmSecs();
    srcsecs = numericvalSecs();
    srcmsecs = numericvalmSecs();;
    secs = srcsecs + addsecs;
    msecs = srcmsecs + addmsecs;
    if (msecs >= 1000) { //Howard found this bug with ESC/Java was ">"
      msecs -= 1000;
      secs++;
    }
    return new returnType(secs,msecs);
  }

  TimePoint minus(TimePoint tp) {
    /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert(false);

    if (eq(tp))
      return (new TimePoint(0,0)); //WV change for nonsymbolic constr
    else if (isPlusInfinity() || tp.isMinusInfinity())
      return (new TimePoint(PlusInfinity));
    else if (isMinusInfinity() || tp.isPlusInfinity())
      return (new TimePoint(MinusInfinity));
    else if (isnumeric() && tp.isnumeric()) {
      returnType rt = numericDiff(tp);
      long secs = rt.getSecs();
      int msecs = rt.getmSecs();
      return (new TimePoint(secs, msecs));
    } else
      return null;
  }

  TimePoint plus(TimePoint tp) {
    /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert(false);

    if ((isMinusInfinity() && tp.isPlusInfinity()) ||
	(isPlusInfinity() && tp.isMinusInfinity()))
      return (new TimePoint(0,0)); // hack again 2nd 0 ignored
    else if (isPlusInfinity() || tp.isPlusInfinity())
      return (new TimePoint(PlusInfinity));
    else if (isMinusInfinity() || tp.isMinusInfinity())
      return (new TimePoint(MinusInfinity));
    else if (isnumeric() && tp.isnumeric()) {
      returnType rt = numericSum(tp);
      long secs = rt.getSecs();
      int msecs = rt.getmSecs();
      return (new TimePoint(secs, msecs));
    } else
      return null;
  }

  void increment(TimePoint tp) {
    /* Corina */
    //if (tp instanceof AbstractTimePoint)
    //      assert(false);

    if ((isMinusInfinity() && tp.isPlusInfinity()) ||
	(isPlusInfinity() && tp.isMinusInfinity()))
      numeric_init(0, 0);
    else if (isPlusInfinity() || tp.isPlusInfinity())
      symbolic_init(PlusInfinity);
    else if (isMinusInfinity() || tp.isMinusInfinity())
      symbolic_init(MinusInfinity);
    else if (isnumeric() && tp.isnumeric()) {
      returnType rt = numericSum(tp);
      long secs = rt.getSecs();
      int msecs = rt.getmSecs();
      numeric_init(secs, msecs);
    } else {
      System.err.println("increment of " + this + " and " + tp + 
	" gives undefined result");
      symbolic_init(SymbolicTimeUndefined);
    }
  }

  boolean lt(TimePoint tp) {
    /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert (false);

    if ((isMinusInfinity() && !tp.isMinusInfinity()) ||
	(tp.isPlusInfinity() && !isPlusInfinity()))
      return true;
    if (!isnumeric() || !tp.isnumeric())
      return false;
    int /*TimePointRelation*/ rel = numeric_relation(tp);
    return (rel == tpLT);
  }
    
  boolean le(TimePoint tp) {
    /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert(false);

    if (isMinusInfinity() || tp.isPlusInfinity())
      return true;
    if (!isnumeric() || !tp.isnumeric())
      return false;
    int /*TimePointRelation*/ rel = numeric_relation(tp);
    return (rel == tpLT || rel == tpEQ);
  }
    
  boolean eq(TimePoint tp) {
    /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert(false);

    return ((isnumeric() && tp.isnumeric() && numeric_equal(tp)) ||
	    (issymbolic() && tp.issymbolic() && symbolic_equal(tp)));
  }
    
  boolean ge(TimePoint tp) {
    /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert(false);

    if (isPlusInfinity() || tp.isMinusInfinity())
      return true;
    if (!isnumeric() || !tp.isnumeric())
      return false;
    int /*TimePointRelation*/ rel = numeric_relation(tp);
    return (rel == tpGT || rel == tpEQ);
  }
    
  boolean gt(TimePoint tp) {
    /* Corina */
    if (tp instanceof AbstractTimePoint)
          assert(false);

    if ((isPlusInfinity() && !tp.isPlusInfinity()) ||
	(tp.isMinusInfinity() && !isMinusInfinity()))
      return true;
    if (!isnumeric() || !tp.isnumeric())
      return false;
    int /*TimePointRelation*/ rel = numeric_relation(tp);
    return (rel == tpGT);
  }

  public int compareTo(Object o)
  {
    TimePoint that = (TimePoint)o;
    if (that instanceof AbstractTimePoint) {
	//WV when comparing we cannot be sure the abstract 
        //   value is on the "left" so we move it there
      return that.compareTo(this);
    }
    if (lt(that)) 
      return -1;
    else if (equal(that))
      return 0;
    else 
      return 1;
    //    return gt(that) ? -1 : equal(that) ? 0 : 1;
  }

  TimePoint roundDown(double discreteSize) {
    if (discreteSize <= 0)
      return (new TimePoint(this));
    if (isMinusInfinity() || isPlusInfinity())
      return (new TimePoint(this));
    else if (isnumeric()) {
      double dval = doubleval();
      double rem = approx_fmod(dval, discreteSize);
      return (new TimePoint(dval - rem));
    } else
      return null;
  }

  TimePoint roundUp(double discreteSize) {
    if (discreteSize <= 0)
      return (new TimePoint(this));
    if (isMinusInfinity() || isPlusInfinity())
      return (new TimePoint(this));
    else if (isnumeric()) {
      double dval = doubleval();
      double rem = approx_fmod(dval, discreteSize);
      if (rem != 0)
	return (new TimePoint(dval + discreteSize - rem));
      else
	return (new TimePoint(dval));
    } else
      return null;
  }  

  // somthing from utils.cc in ....crl-def.../

  double approx_fmod(double number, double modulus) {
    int fmod_multiplier = 100000;
    int mult_number = (int) (number * fmod_multiplier + 0.5);
    int mult_mod = (int) (modulus * fmod_multiplier + 0.5);
    int mult_rem;
    if (mult_mod == 0)
      return 0.0;
    mult_rem = mult_number % mult_mod;
    return ((double) mult_rem / fmod_multiplier);
  }

}




